import base64

import pyasice

from ..base_service import BaseSKService
from ..constants import HASH_ALGORITHMS, HASH_SHA256
from ..exceptions import CanceledByUser
from ..exceptions import EsteidError as MobileIDError
from ..exceptions import InvalidIdCode, SignatureVerificationError, UpstreamServiceError, UserNotRegistered, UserTimeout
from ..util import generate_hash, id_code_ee_is_valid, secure_random
from .constants import EndResults, Languages
from .types import AuthenticateResult, AuthenticateStatusResult, SignResult, SignStatusResult
from .utils import get_verification_code


class MobileIDService(BaseSKService):
    """Mobile-ID Authentication and signing

    Based on https://github.com/SK-EID/MID

    Test api root: https://tsp.demo.sk.ee/mid-api
    Production api root: https://mid.sk.ee/mid-api

    Fixme:
        - Add more logging
    """

    # Notes: the texts must not exceed 20 characters in length!
    DISPLAY_TEXT_AUTH = "Authenticate"
    DISPLAY_TEXT_SIGN = "Sign"

    NAME = "Mobile ID"

    class Actions:
        AUTH = "/authentication"
        SIGN = "/signature"
        SESSION_STATUS = "{action}/session/{session_id}"

    def get_certificate(self, id_code: str, phone_number: str) -> bytes:
        """
        Gets a user certificate that would be used for signing.

        Also checks if the user with these
        """
        if not (id_code and phone_number):
            # TODO proper validation
            raise ValueError("Both id_code and phone_number are required")

        if not id_code_ee_is_valid(id_code):
            raise InvalidIdCode

        endpoint = "/certificate"

        result = self.invoke(
            endpoint,
            method="POST",
            data={"phoneNumber": phone_number, "nationalIdentityNumber": id_code},
        )

        status = result["result"]
        if status == "OK":
            return base64.b64decode(result["cert"])
        if status == "NOT_FOUND":
            raise UserNotRegistered(
                "User with phone number {phone} and ID code {id_code} not found".format(
                    phone=phone_number, id_code=id_code
                )
            )
        raise MobileIDError("Unknown response format")

    def authenticate(
        self, id_code: str, phone_number: str, message: str = None, language: str = None, hash_type: str = HASH_SHA256
    ):
        """Initiate an authentication session

        https://github.com/SK-EID/MID#32-initiating-signing-and-authentication

        Difference between auth and signing is the hash: for auth, it's just a random string of the corresponding size
        (the size of a hash produced by selected algorithm) - 32 for SHA256

        https://github.com/SK-EID/MID#23-creating-the-hash

        :param str id_code: National identity number
        :param str phone_number: User's phone number
        :param str message: Text to display for authentication consent dialog on the mobile device, NOTE: max 20 chars!
        :param str language: choices: ENG, EST, LIT, RUS. Defaults to ENG
        :param str hash_type: Hash algorithm to use when generating a random hash value (choices: HASH_ALGORITHMS)
        :return: session ID
        :rtype: AuthenticateResult
        """
        # Ensure required values are set
        if not (id_code and phone_number):
            # TODO proper validation
            raise ValueError("Both id_code and phone_number are required")

        if not id_code_ee_is_valid(id_code):
            raise InvalidIdCode

        assert hash_type in HASH_ALGORITHMS

        message = str(message or self.DISPLAY_TEXT_AUTH)
        assert len(message) <= 20, f"Display text can not exceed 20 chars: got '{message}'"

        if language not in Languages.ALL:
            language = Languages.ENG

        random_bytes = secure_random(64)
        hash_value = generate_hash(hash_type, random_bytes)
        hash_value_b64 = base64.b64encode(hash_value)

        endpoint = self.Actions.AUTH

        result = self.invoke(
            endpoint,
            method="POST",
            data={
                "nationalIdentityNumber": id_code,
                "phoneNumber": phone_number,
                "hashType": hash_type,
                "hash": hash_value_b64.decode("utf-8"),
                "language": language,
                # Casting to str to ensure translations are resolved
                "displayText": message,  # NOTE: hard 20-char limit
                "displayTextFormat": "UCS-2",  # the other choice is GSM-7 which is 7-bit
            },
        )

        assert "sessionID" in result, "No session id in {result}".format(result=result)
        return AuthenticateResult(
            session_id=result["sessionID"],
            digest=hash_value,
            hash_type=hash_type,
            verification_code=get_verification_code(hash_value),
        )

    def status(self, session_id: str, digest: bytes, timeout: int = 10000):
        """
        Retrieve auth session result from Mobile-ID backend

        see https://github.com/SK-EID/MID#33-status-of-signing-and-authentication

        :param session_id: session ID, from auth result
        :param digest: hash value that was used to initiate auth
        :param int timeout: Request long poll timeout value in milliseconds (Note: server uses a default
                         if client does not send it)
        :rtype: AuthenticateStatusResult
        """
        data = self._get_session_response(self.Actions.AUTH, session_id, timeout)

        # signature.value (str): Signature value, base64 encoded.
        signature_value = base64.b64decode(data["signature"]["value"])

        # signature.algorithm (str): Signature algorithm, in the form of sha256WithRSAEncryption
        signature_algorithm = data["signature"]["algorithm"]
        assert signature_algorithm[:6].upper() in HASH_ALGORITHMS

        # cert: Certificate value, DER+Base64 encoded
        cert_value = base64.b64decode(data["cert"])

        try:
            pyasice.verify(cert_value, signature_value, digest, signature_algorithm[:6], prehashed=True)
        except pyasice.SignatureVerificationError as e:
            raise SignatureVerificationError from e

        return AuthenticateStatusResult(
            certificate=cert_value,
            signature=signature_value,
            signature_algorithm=signature_algorithm,
        )

    def sign(
        self,
        id_code: str,
        phone_number: str,
        signed_data: bytes,
        message: str = None,
        language: str = None,
        hash_type: str = HASH_SHA256,
    ):
        """Initiate a signature session.

        Differences from authentication session:
        - endpoint
        - using signed data instead of a random string

        https://github.com/SK-EID/MID#32-initiating-signing-and-authentication

        :param str id_code: National identity number
        :param str phone_number: Client's phone number
        :param bytes signed_data: Binary data to sign
        :param str message: Text to display on the mobile device, NOTE: max 20 chars!
        :param str language: choices: ENG, EST, LIT, RUS. Defaults to ENG
        :param str hash_type: Hash algorithm used to sign data
        :return SignResult: Result of the request
        """
        if not (id_code and phone_number):
            # TODO proper validation
            raise ValueError("Both id_code and phone_number are required")

        if not id_code_ee_is_valid(id_code):
            raise InvalidIdCode

        hash_type = hash_type.upper()
        assert hash_type in HASH_ALGORITHMS

        message = str(message or self.DISPLAY_TEXT_SIGN)
        assert len(message) <= 20, f"Display text can not exceed 20 chars: got '{message}'"

        if language not in Languages.ALL:
            language = Languages.ENG

        content_hash = generate_hash(hash_type, signed_data)
        hash_value_b64 = base64.b64encode(content_hash)

        result = self.invoke(
            self.Actions.SIGN,
            method="POST",
            data={
                "nationalIdentityNumber": id_code,
                "phoneNumber": phone_number,
                "hashType": hash_type,
                "hash": hash_value_b64.decode("utf-8"),
                "language": language,
                # Casting to str to ensure translations are resolved
                "displayText": message,
                "displayTextFormat": "UCS-2",  # the other choice is GSM-7 which is 7-bit
            },
        )

        return SignResult(
            session_id=result["sessionID"],
            digest=content_hash,
            verification_code=get_verification_code(content_hash),
        )

    def sign_status(self, session_id: str, certificate: bytes, signed_digest: bytes, timeout=10000):
        """Retrieve signing session result from Mobile-ID backend

        The certificate is missing in the response, so
        to verify the signature you must pass it on from the authentication process.

        see https://github.com/SK-EID/MID#33-status-of-signing-and-authentication

        :param session_id: session ID from I{sign} Result
        :param certificate: the certificate used to sign data
        :param signed_digest: the digest of data to sign
        :param int timeout: Request long poll timeout value in milliseconds (Note: server uses a default
                         if client does not send it)
        :rtype: SignStatusResult
        """
        data = self._get_session_response(self.Actions.SIGN, session_id, timeout)

        signature = base64.b64decode(data["signature"]["value"])
        signature_algorithm = data["signature"]["algorithm"]
        assert signature_algorithm[:6].upper() in HASH_ALGORITHMS

        try:
            pyasice.verify(certificate, signature, signed_digest, signature_algorithm[:6], prehashed=True)
        except pyasice.SignatureVerificationError as e:
            raise SignatureVerificationError from e

        return SignStatusResult(
            signature=signature,
            signature_algorithm=signature_algorithm,
            certificate=certificate,
        )

    # ============
    # Internals
    # ============

    def _get_session_response(self, action, session_id, timeout=None):
        """Perform a request to session status.

        :param action: authentication or signature
        :param session_id:
        :param timeout:
        :return:
        """
        endpoint = self.Actions.SESSION_STATUS.format(action=action, session_id=session_id)

        data = self.poll_session(session_id, endpoint_url=endpoint, timeout=timeout)

        # result (str): End result of the transaction.
        # This structure is different from Smart ID so can't efficiently put it into BaseSKService.poll_session()
        end_result = data["result"]

        if end_result != EndResults.OK:
            if end_result == EndResults.TIMEOUT:
                raise UserTimeout
            elif end_result == EndResults.USER_CANCELLED:
                raise CanceledByUser
            elif end_result == EndResults.NOT_MID_CLIENT:
                raise UserNotRegistered
            # Fail hard, if endResult is something unknown to us
            if end_result not in EndResults.ALL:
                raise MobileIDError(f"Unexpected result '{end_result}' reported")

            raise UpstreamServiceError(f"Service returned {end_result}", service=self.NAME)

        return data



import binascii
from unittest.mock import Mock, patch

import pytest

from esteid.exceptions import InvalidParameter
from esteid.idcard import IdCardSigner


@pytest.fixture()
def idcardsigner():
    signer = IdCardSigner({}, initial=True)
    mock_container = Mock(name="mock_container")
    with patch.object(signer, "open_container", return_value=mock_container), patch.object(signer, "save_session_data"):
        mock_container.prepare_signature.return_value = mock_xml_sig = Mock(name="mock_xml_sig")
        mock_xml_sig.digest.return_value = b"some binary digest"

        yield signer


def test_idcardsigner_certificate(static_certificate):
    signer = IdCardSigner({}, initial=True)
    signer.setup({"certificate": binascii.b2a_hex(static_certificate)})

    assert signer.id_code == "11702020200"

    with pytest.raises(InvalidParameter, match="Missing required parameter 'certificate'"):
        signer.setup({})

    with pytest.raises(InvalidParameter, match="Failed to decode"):
        signer.setup({"certificate": b"something not hex-encoded"})

    with pytest.raises(InvalidParameter, match="supported certificate format"):
        signer.setup({"certificate": binascii.b2a_hex(b"this is not a certificate")})


def test_idcardsigner_prepare(idcardsigner, static_certificate):
    idcardsigner.setup({"certificate": binascii.b2a_hex(static_certificate)})

    result = idcardsigner.prepare(None, [])

    idcardsigner.open_container.assert_called_once_with(None, [])
    container = idcardsigner.open_container(...)

    container.prepare_signature.assert_called_once_with(static_certificate)

    xml_sig = idcardsigner.open_container().prepare_signature(...)
    assert xml_sig._mock_name == "mock_xml_sig"

    idcardsigner.save_session_data.assert_called_once_with(
        digest=xml_sig.digest(),
        container=container,
        xml_sig=xml_sig,
    )

    assert result == {"digest": binascii.b2a_hex(xml_sig.digest()).decode()}



import pytest

import requests_mock
from django.core.exceptions import ImproperlyConfigured
from requests.exceptions import ConnectTimeout

from ...constants import MOBILE_ID_DEMO_URL, MOBILE_ID_LIVE_URL
from ...exceptions import EsteidError, InvalidCredentials, OfflineError, UpstreamServiceError
from .. import MobileIDError
from ..base import MobileIDService
from ..i18n import TranslatedMobileIDService


def test_error():
    assert MobileIDError is EsteidError


@pytest.mark.parametrize("exc", [ConnectionError, ConnectTimeout])
def test_mobileid_invoke_timeout(demo_mid_api, exc):
    with requests_mock.mock() as m:
        m.get(demo_mid_api.api_url(""), exc=exc)

        with pytest.raises(OfflineError) as exc_info:
            demo_mid_api.invoke("")

        assert demo_mid_api.NAME in exc_info.value.get_message()


@pytest.mark.parametrize(
    "status_code,exc",
    [
        (401, InvalidCredentials),
        (580, OfflineError),
        (502, OfflineError),
        (503, OfflineError),
        (504, OfflineError),
        (400, MobileIDError),
        (500, UpstreamServiceError),
    ],
)
def test_mobileid_invoke_errors(demo_mid_api, status_code, exc):
    with requests_mock.mock() as m:
        m.get(demo_mid_api.api_url(""), status_code=status_code)

        with pytest.raises(exc):
            demo_mid_api.invoke("")


def test_mobileid_service(demo_mid_api):
    assert demo_mid_api.api_root == MOBILE_ID_DEMO_URL

    service = MobileIDService("00000000-0000-0000-0000-000000000000", "test", MOBILE_ID_LIVE_URL)
    assert service.api_root == MOBILE_ID_LIVE_URL
    assert service.rp_uuid == "00000000-0000-0000-0000-000000000000"
    assert service.rp_name == "test"


def test_mobileid_translated_service(i18n_demo_mid_api):
    assert i18n_demo_mid_api.api_root == MOBILE_ID_DEMO_URL


def test_mobileid_translated_service_test_mode_off(i18n_demo_mid_api, override_esteid_settings):
    with override_esteid_settings(
        MOBILE_ID_TEST_MODE=False,
        MOBILE_ID_SERVICE_UUID="00000000-0000-0000-0000-000000000000",
        MOBILE_ID_SERVICE_NAME="test",
    ):
        service = TranslatedMobileIDService.get_instance()
        assert service.api_root == MOBILE_ID_LIVE_URL
        assert service.rp_uuid == "00000000-0000-0000-0000-000000000000"
        assert service.rp_name == "test"


def test_mobileid_translated_service_requires_creds_for_live(override_esteid_settings):
    with override_esteid_settings(MOBILE_ID_TEST_MODE=False):
        with pytest.raises(ImproperlyConfigured, match="MOBILE_ID_SERVICE_NAME and MOBILE_ID_SERVICE_UUID"):
            TranslatedMobileIDService.get_instance()

    with override_esteid_settings(
        MOBILE_ID_TEST_MODE=False,
        MOBILE_ID_SERVICE_UUID="00000000-0000-0000-0000-000000000000",
    ):
        with pytest.raises(ImproperlyConfigured, match="MOBILE_ID_SERVICE_NAME and MOBILE_ID_SERVICE_UUID"):
            TranslatedMobileIDService.get_instance()

    with override_esteid_settings(MOBILE_ID_TEST_MODE=False, MOBILE_ID_SERVICE_NAME="name"):
        with pytest.raises(ImproperlyConfigured, match="MOBILE_ID_SERVICE_NAME and MOBILE_ID_SERVICE_UUID"):
            TranslatedMobileIDService.get_instance()



import pytest

from esteid.constants import HASH_SHA256
from esteid.smartid.utils import get_verification_code
from esteid.util import generate_hash


@pytest.mark.parametrize(
    "expected_verification_code,hash_raw",
    [
        ("7712", b"Hello World!"),
        ("3404", b"You broke it, didn't you."),
        ("0914", b"Weeeeeeeeeeeeeeeeeeeeee[bzzt]"),
    ],
)
def test_verification_code_generator(expected_verification_code, hash_raw):
    assert get_verification_code(generate_hash(HASH_SHA256, hash_raw)) == expected_verification_code



import hashlib

import pytest

from esteid.constants import HASH_ALGORITHMS
from esteid.util import generate_hash, id_code_ee_is_valid, id_code_lt_is_valid, id_code_lv_is_valid


@pytest.mark.parametrize("hash_algo", HASH_ALGORITHMS)
def test_hashtypes(hash_algo):
    hash_value = generate_hash(hash_algo, b"FOOBAR")

    hash_method = getattr(hashlib, hash_algo.lower())
    assert hash_value == hash_method(b"FOOBAR").digest()


def test_invalid_hash_algo_fails():
    with pytest.raises(ValueError):
        generate_hash("FAKEHASH", b"FOOBAR")


@pytest.mark.parametrize(
    "id_code, result",
    [
        pytest.param("", False, id="empty string"),
        pytest.param(None, False, id="None"),
        pytest.param(["10101010005"], False, id="list"),
        pytest.param("ABCDEFGHQWE", False, id="not digits"),
        pytest.param("10101010", False, id="not enough digits"),
        pytest.param("10101010OO5", False, id="not all digits (0-O)"),
        pytest.param("123456789O123456", False, id="too many digits"),
        pytest.param("10101010101", False, id="wrong checksum"),
        pytest.param("10101010005", True, id="test SmartID code"),
        pytest.param("60001019906", True, id="test MobileID code"),
        pytest.param("37605030299", True, id="From Wikipedia"),
    ],
)
def test_id_code_ee_lt_is_valid(id_code, result):
    assert id_code_ee_is_valid(id_code) == result
    assert id_code_lt_is_valid(id_code) == result, "LT ID codes are same format as EE"


@pytest.mark.parametrize(
    "id_code, result",
    [
        pytest.param("", False, id="empty string"),
        pytest.param(None, False, id="None"),
        pytest.param(["010101-10006"], False, id="list"),
        pytest.param("ABCDEF-QWERT", False, id="not digits"),
        pytest.param("01010110006", False, id="no dash"),
        pytest.param("010101-1OOO6", False, id="not all digits (0-O)"),
        pytest.param("123456-789O123456", False, id="too many digits"),
        pytest.param("010101-01010", False, id="wrong checksum"),
        pytest.param("010101-10006", True, id="test MobileID code"),
        pytest.param("111111-11111", True, id="From gihtub/id-lv"),
    ],
)
def test_id_code_lv_is_valid(id_code, result):
    assert id_code_lv_is_valid(id_code) == result



import binascii
import logging
import typing
import warnings
from tempfile import NamedTemporaryFile

from esteid_certificates import get_certificate

import pyasice
from pyasice import Container, finalize_signature, verify, XmlSignature

from esteid import settings
from esteid.exceptions import ActionInProgress, EsteidError, InvalidIdCode, UserNotRegistered
from esteid.mobileid.i18n import TranslatedMobileIDService
from esteid.smartid.i18n import TranslatedSmartIDService

from .session import delete_esteid_session, get_esteid_session, open_container, update_esteid_session
from .util import id_code_ee_is_valid


if typing.TYPE_CHECKING:
    from .generic import GenericDigitalSignViewMixin

warnings.warn("The actions API is deprecated. Please use the new signing API", DeprecationWarning)

logger = logging.getLogger(__name__)


class BaseAction(object):
    @classmethod
    def do_action(cls, view: "GenericDigitalSignViewMixin", params):
        raise NotImplementedError


class NoAction(object):
    @classmethod
    def do_action(cls, view, params):
        return {"success": True}


class IdCardPrepareAction(BaseAction):
    @classmethod
    def do_action(cls, view: "GenericDigitalSignViewMixin", params: dict = None, *, certificate: str = None):
        """
        The old API is to pass a dict of params (previously confusingly named `action_kwargs`),
        the keyword args are added here for clarity as to what the method accepts

        :param view:
        :param params:
        :param certificate: HEX-encoded certificate from the ID card
        :return:
        """
        request = view.request
        delete_esteid_session(request)

        if not certificate:
            certificate = (params or {}).get("certificate")

        if not certificate:
            return {
                "success": False,
                "code": "BAD_CERTIFICATE",
            }

        certificate = binascii.a2b_hex(certificate)

        files = view.get_files()
        container_path = view.get_bdoc_container_file()

        if not (files or container_path):
            return {
                "success": False,
                "code": "MIN_1_FILE",
            }

        container = open_container(container_path, files)
        xml_sig = container.prepare_signature(certificate)

        # save intermediate signature XML to temp file
        with NamedTemporaryFile(delete=False) as temp_signature_file:
            temp_signature_file.write(xml_sig.dump())

        # always save container to a temp file
        with NamedTemporaryFile(mode="wb", delete=False) as temp_container_file:
            temp_container_file.write(container.finalize().getbuffer())

        signed_digest = xml_sig.digest()
        digest_hash_b64 = binascii.b2a_base64(signed_digest).decode()

        update_esteid_session(
            request,
            signed_hash=digest_hash_b64,  # we can take the hash from the signature XML, but it'd take time to compute
            temp_signature_file=temp_signature_file.name,
            temp_container_file=temp_container_file.name,
        )

        return {
            "success": True,
            "digest": binascii.b2a_hex(signed_digest).decode(),
        }


class IdCardFinishAction(BaseAction):
    @classmethod
    def do_action(cls, view: "GenericDigitalSignViewMixin", params: dict = None, *, signature_value: str = None):
        """
        The old API is to pass a dict of params (previously confusingly named `action_kwargs`),
        the keyword args are added here for clarity as to what the method accepts

        :param view:
        :param params:
        :param signature_value: a HEX encoded signature, as received from `hwcrypto.js`
        :return:
        """
        request = view.request
        session_data = get_esteid_session(request)
        if not session_data:
            return {
                "success": False,
                "code": "NO_SESSION",
            }

        if signature_value is None:
            signature_value = (params or {}).get("signature_value")
            if not signature_value:
                return {
                    "success": False,
                    "code": "BAD_SIGNATURE",
                }

        logger.debug("Signature HEX: %s", signature_value)

        signed_hash_b64 = session_data["signed_hash"]
        signature_value = binascii.a2b_hex(signature_value)

        temp_signature_file = session_data["temp_signature_file"]
        temp_container_file = session_data["temp_container_file"]

        with open(temp_signature_file, "rb") as f:
            xml_sig = XmlSignature(f.read())

        # Load a partially prepared BDoc from a tempfile and clean it up
        container = Container.open(temp_container_file)

        # now we don't need the session anymore
        delete_esteid_session(request)

        verify(xml_sig.get_certificate_value(), signature_value, binascii.a2b_base64(signed_hash_b64), prehashed=True)

        xml_sig.set_signature_value(signature_value)
        issuer_cert = get_certificate(xml_sig.get_certificate_issuer_common_name())

        try:
            finalize_signature(
                xml_sig,
                issuer_cert,
                lt_ts=settings.ESTEID_USE_LT_TS,
                ocsp_url=settings.OCSP_URL,
                tsa_url=settings.TSA_URL,
            )
        except pyasice.Error:
            logger.exception("Signature confirmation service error")
            return {
                "success": False,
                "code": "SERVICE_ERROR",
            }

        container.add_signature(xml_sig)

        return {
            "success": True,
            "container": container,
        }


class SignCompleteAction(BaseAction):
    @classmethod
    def do_action(cls, view, action_kwargs):
        """Return the signed container"""
        raise NotImplementedError


class MobileIdSignAction(BaseAction):
    @classmethod
    def do_action(
        cls,
        view: "GenericDigitalSignViewMixin",
        params: dict = None,
        *,
        phone_number: str = None,
        id_code: str = None,
        language: str = None,
    ):
        """
        The old API is to pass a dict of params (previously confusingly named `action_kwargs`),
        the keyword args are added here for clarity as to what the method accepts

        :param view:
        :param params:
        :param phone_number:
        :param id_code:
        :param language:
        :return:
        """
        request = view.request
        delete_esteid_session(request)

        if not phone_number:
            phone_number = params["phone_number"]
        if not id_code:
            id_code = params["id_code"]

        if not (phone_number and id_code):
            return {
                "success": False,
                "code": "BAD_PARAMS",
            }

        # NOTE: since EE and LT id codes use the same format, we are using the same function.
        if not id_code_ee_is_valid(id_code):
            return {
                "success": False,
                "code": "INVALID_ID_CODE",
            }

        files = view.get_files()
        container_path = view.get_bdoc_container_file()

        if not (files or container_path):
            return {
                "success": False,
                "code": "MIN_1_FILE",
            }

        service = TranslatedMobileIDService.get_instance()

        try:
            certificate = service.get_certificate(id_code, phone_number)
        except UserNotRegistered:
            return {
                "success": False,
                "code": "NOT_A_MOBILEID_USER",
            }

        container = open_container(container_path, files)
        xml_sig = container.prepare_signature(certificate)

        # save intermediate signature XML to temp file
        with NamedTemporaryFile(delete=False) as temp_signature_file:
            temp_signature_file.write(xml_sig.dump())

        # always save container to a temp file
        with NamedTemporaryFile(mode="wb", delete=False) as temp_container_file:
            temp_container_file.write(container.finalize().getbuffer())

        try:
            sign_session = service.sign(id_code, phone_number, xml_sig.signed_data(), language=language)
        except EsteidError:
            return {
                "success": False,
                "code": "SIGN_SESSION_FAILED",
            }

        signed_digest = sign_session.digest
        digest_hash_b64 = binascii.b2a_base64(signed_digest).decode()

        update_esteid_session(
            request,
            session_id=sign_session.session_id,
            digest_b64=digest_hash_b64,
            temp_signature_file=temp_signature_file.name,
            temp_container_file=temp_container_file.name,
        )

        return {
            "success": True,
            "challenge": sign_session.verification_code,
            "verification_code": sign_session.verification_code,
        }


class MobileIdStatusAction(BaseAction):
    @classmethod
    def do_action(cls, view: "GenericDigitalSignViewMixin", params: dict = None):

        request = view.request
        session_data = get_esteid_session(request)
        if not session_data:
            return {
                "success": False,
                "code": "NO_SESSION",
            }

        session_id = session_data["session_id"]
        signed_digest = session_data["digest_b64"]
        temp_signature_file = session_data["temp_signature_file"]
        temp_container_file = session_data["temp_container_file"]

        with open(temp_signature_file, "rb") as f:
            xml_sig = XmlSignature(f.read())

        # Load a partially prepared BDoc from a tempfile and clean it up
        container = Container.open(temp_container_file)

        service = TranslatedMobileIDService.get_instance()
        try:
            status = service.sign_status(
                session_id, xml_sig.get_certificate_value(), binascii.a2b_base64(signed_digest)
            )
        except ActionInProgress:
            #  when there is an `ActionInProgress` exception, we shouldn't delete the session.
            return {
                "success": False,
                "pending": True,
            }
        except Exception:
            # NOTE: we could pick some exceptions that don't require cleanup,
            # but this also requires support from the party that polls this action.
            # Most likely the whole process would need to be restarted anyway
            delete_esteid_session(request)
            raise

        # now we don't need the session anymore
        delete_esteid_session(request)

        xml_sig.set_signature_value(status.signature)

        issuer_cert = get_certificate(xml_sig.get_certificate_issuer_common_name())

        try:
            finalize_signature(
                xml_sig,
                issuer_cert,
                lt_ts=settings.ESTEID_USE_LT_TS,
                ocsp_url=settings.OCSP_URL,
                tsa_url=settings.TSA_URL,
            )
        except pyasice.Error:
            logger.exception("Signature confirmation service error")
            return {
                "success": False,
                "code": "SERVICE_ERROR",
            }

        container.add_signature(xml_sig)

        return {
            "success": True,
            "container": container,
        }


class SmartIdSignAction(BaseAction):
    @classmethod
    def do_action(
        cls,
        view: "GenericDigitalSignViewMixin",
        params: dict = None,
        *,
        id_code: str = None,
        country: str = None,
        language: str = None,
    ):
        """
        The old API is to pass a dict of params (previously confusingly named `action_kwargs`),
        the keyword args are added here for clarity as to what the method accepts
        """
        request = view.request
        delete_esteid_session(request)

        params = params or {}
        if not id_code:
            id_code = params.get("id_code")

        if not country:
            country = params.get("country") or settings.ESTEID_COUNTRY

        files = view.get_files()
        container_path = view.get_bdoc_container_file()

        if not (files or container_path):
            return {
                "success": False,
                "code": "MIN_1_FILE",
            }

        service = TranslatedSmartIDService.get_instance()

        try:
            certificate, document_number = service.select_signing_certificate(id_code=id_code, country=country)
        except InvalidIdCode:
            return {
                "success": False,
                "code": "INVALID_ID_CODE",
            }
        except UserNotRegistered:
            return {
                "success": False,
                "code": "NOT_A_SMARTID_USER",
            }
        except EsteidError:
            logger.exception("An error occurred during SmartID certificate selection")
            raise

        container = open_container(container_path, files)
        xml_sig = container.prepare_signature(certificate)

        sign_result = service.sign_by_document_number(document_number, xml_sig.signed_data())

        signed_digest = sign_result.digest
        digest_hash_b64 = binascii.b2a_base64(signed_digest).decode()

        # always save container to a temp file
        with NamedTemporaryFile(mode="wb", delete=False) as temp_container_file:
            temp_container_file.write(container.finalize().getbuffer())

        # save intermediate signature XML to temp file
        with NamedTemporaryFile(delete=False) as temp_signature_file:
            temp_signature_file.write(xml_sig.dump())

        update_esteid_session(
            request,
            session_id=sign_result.session_id,
            digest_b64=digest_hash_b64,
            temp_container_file=temp_container_file.name,
            temp_signature_file=temp_signature_file.name,
        )

        return {
            "success": True,
            "verification_code": sign_result.verification_code,
        }


class SmartIdStatusAction(BaseAction):
    @classmethod
    def do_action(cls, view: "GenericDigitalSignViewMixin", params: dict = None):
        request = view.request
        session_data = get_esteid_session(request)
        if not session_data:
            return {
                "success": False,
                "code": "NO_SESSION",
            }

        # Continue with signing - poll status and finalize
        logger.debug("SmartID: polling status of signing")

        session_id = session_data["session_id"]
        signed_digest = session_data["digest_b64"]
        temp_signature_file = session_data["temp_signature_file"]
        temp_container_file = session_data["temp_container_file"]

        service = TranslatedSmartIDService.get_instance()
        try:
            status = service.sign_status(session_id, binascii.a2b_base64(signed_digest))
        except ActionInProgress:
            # Do not delete session here.
            return {
                "success": False,
                "pending": True,
            }
        except Exception:
            logger.exception("Failed to get signing status from Smart ID service")
            # NOTE: we could pick some exceptions that don't require cleanup,
            # but this also requires support from the party that polls this action.
            # Most likely the whole process would need to be restarted anyway
            delete_esteid_session(request)
            raise

        logger.debug("SmartID Signing complete")

        with open(temp_signature_file, "rb") as f:
            xml_sig = XmlSignature(f.read())

        # Load a partially prepared BDoc from a tempfile and clean it up
        container = Container.open(temp_container_file)

        # now we don't need the session anymore
        delete_esteid_session(request)

        xml_sig.set_signature_value(status.signature)

        issuer_cert = get_certificate(xml_sig.get_certificate_issuer_common_name())

        try:
            finalize_signature(
                xml_sig,
                issuer_cert,
                lt_ts=settings.ESTEID_USE_LT_TS,
                ocsp_url=settings.OCSP_URL,
                tsa_url=settings.TSA_URL,
            )
        except pyasice.Error:
            logger.exception("Signature confirmation service error")
            return {
                "success": False,
                "code": "SERVICE_ERROR",
            }

        container.add_signature(xml_sig)

        return {
            "success": True,
            "container": container,
        }



import logging
import os
import subprocess
import warnings
from tempfile import NamedTemporaryFile

from django.utils.encoding import force_bytes, force_text
from esteid_certificates import get_certificate_file_name, UnknownCertificateError


warnings.warn("This module is deprecated. Please use the new signing API", DeprecationWarning)


logger = logging.getLogger(__name__)


class OCSPVerifier(object):
    def __init__(self, certificate, issuer, ocsp_url, va_file_path):
        self.certificate = certificate
        self.issuer = issuer
        if isinstance(self.issuer, str):
            self.issuer = self.parse_issuer(issuer)

        self.ocsp_url = ocsp_url
        self.va_file_path = va_file_path

    def verify(self):
        if not self.certificate:
            return 1

        if not self.issuer:
            return 2

        issuer_cn = self.issuer.get("CN", None)
        if not issuer_cn:
            return 3

        try:
            issuer_cert_filename = get_certificate_file_name(issuer_cn)
        except UnknownCertificateError:
            return 4

        # Save cert to a temporary file
        cert_file = NamedTemporaryFile(delete=False)
        cert_file.write(force_bytes(self.certificate.replace("\t", "")))
        cert_file.close()

        # Run openssl ocsp verify command
        args = [
            "openssl",
            "ocsp",
            "-issuer %s" % issuer_cert_filename,
            "-cert %s" % cert_file.name,
            "-url %s" % self.ocsp_url,
            "-VAfile %s" % self.va_file_path,
        ]

        try:
            output = subprocess.check_output(" ".join(args), stderr=subprocess.STDOUT, shell=True)
            output = force_text(output)

            if ": good" in output:
                result = 0

            elif ": revoked" in output:
                result = 6

            else:
                logger.info("openssl ocsp: unknown output: %s", output)
                result = 7

        except subprocess.CalledProcessError as e:
            logger.error("openssl ocsp: failed with output: %s", e.output)
            result = 5

        # delete temporary file
        os.unlink(cert_file.name)

        return result

    @classmethod
    def parse_issuer(cls, issuer):
        res = {}
        issuer = issuer.strip().strip("/").split("/")

        for part in issuer:
            part = part.split("=")
            res[part[0]] = part[1]

        return res



MOBILE_ID_DEMO_URL = "https://tsp.demo.sk.ee/mid-api"
MOBILE_ID_LIVE_URL = "https://mid.sk.ee/mid-api"

MOBILE_ID_DEMO_SERVICE_NAME = "DEMO"
MOBILE_ID_DEMO_SERVICE_UUID = "00000000-0000-0000-0000-000000000000"

SMART_ID_DEMO_URL = "https://sid.demo.sk.ee/smart-id-rp/v1"
SMART_ID_LIVE_URL = "https://rp-api.smart-id.com/v1"

SMART_ID_DEMO_SERVICE_NAME = MOBILE_ID_DEMO_SERVICE_NAME
SMART_ID_DEMO_SERVICE_UUID = MOBILE_ID_DEMO_SERVICE_UUID

OCSP_DEMO_URL = "http://demo.sk.ee/ocsp"
OCSP_LIVE_URL = "http://ocsp.sk.ee/"

TSA_DEMO_URL = "http://demo.sk.ee/tsa/"
TSA_LIVE_URL = "http://tsa.sk.ee"

# Set of supported hash algorithms. Same for both MobileID and SmartID
# see https://github.com/SK-EID/smart-id-documentation#33-hash-algorithms
# https://github.com/SK-EID/MID#231-supported-hashing-algorithms
HASH_SHA256 = "SHA256"
HASH_SHA384 = "SHA384"
HASH_SHA512 = "SHA512"
HASH_ALGORITHMS = {
    HASH_SHA256,
    HASH_SHA384,
    HASH_SHA512,
}


class Countries:
    ESTONIA = "EE"
    LATVIA = "LV"
    LITHUANIA = "LT"

    ALL = (ESTONIA, LATVIA, LITHUANIA)



from time import sleep, time
from unittest.mock import patch

import pytest

import esteid_certificates
import requests_mock

from pyasice import Container, XmlSignature
from pyasice.ocsp import OCSP
from pyasice.tsa import TSA

from esteid.constants import HASH_ALGORITHMS, HASH_SHA256, HASH_SHA384, HASH_SHA512, OCSP_DEMO_URL, TSA_DEMO_URL

from ...exceptions import (
    ActionInProgress,
    BadRequest,
    InvalidCredentials,
    InvalidIdCode,
    PermissionDenied,
    UserNotRegistered,
)
from ...util import generate_hash
from ..base import SmartIDService
from ..constants import CERTIFICATE_LEVEL_QUALIFIED, Countries
from ..types import AuthenticateResult, AuthenticateStatusResult, SignResult, SignStatusResult


@pytest.mark.parametrize("hash_type", HASH_ALGORITHMS)
def test_authentication(demo_api, hash_type, SMARTID_DEMO_ID_CODE_EE):
    raw_data = b"Hello World!"
    response_data = {"sessionID": "FAKE"}
    verification_codes = {
        HASH_SHA256: "7712",
        HASH_SHA384: "3486",
        HASH_SHA512: "4664",
    }

    known_hashes = {
        HASH_SHA256: b"\x7f\x83\xb1e\x7f\xf1\xfcS\xb9-\xc1\x81H\xa1\xd6]\xfc-K\x1f\xa3\xd6w(J\xdd\xd2\x00\x12m\x90i",
        HASH_SHA384: b"\xbf\xd7l\x0e\xbb\xd0\x06\xfe\xe5\x83A\x05G\xc1\x88{\x02\x92\xbev\xd5\x82\xd9l$-*y'#\xe3"
        b"\xfdo\xd0a\xf9\xd5\xcf\xd1;\x8f\x96\x13X\xe6\xad\xbaJ",
        HASH_SHA512: b"\x86\x18D\xd6pN\x85s\xfe\xc3M\x96~ \xbc\xfe\xf3\xd4$\xcfH\xbe\x04\xe6\xdc\x08\xf2\xbdX\xc7)t3q"
        b"\x01^\xad\x89\x1c\xc3\xcf\x1c\x9d4\xb4\x92d\xb5\x10u\x1b\x1f\xf9\xe57\x93{\xc4k]o\xf4\xec\xc8",
    }

    with patch("esteid.smartid.base.secure_random", return_value=raw_data):
        with patch("esteid.smartid.base.generate_hash", return_value=known_hashes[hash_type]) as mock:
            with patch.object(demo_api, "invoke", return_value=response_data):
                res = demo_api.authenticate(SMARTID_DEMO_ID_CODE_EE, Countries.ESTONIA, hash_type=hash_type)

                mock.assert_called_with(hash_type, raw_data)

                assert isinstance(res, AuthenticateResult)
                assert res.session_id == "FAKE"
                assert res.hash_type == hash_type
                assert res.hash_value == generate_hash(hash_type, raw_data)
                assert res.verification_code == verification_codes[hash_type]


@pytest.mark.parametrize(
    "status,exc",
    [
        pytest.param(400, BadRequest),
        pytest.param(401, InvalidCredentials),
        pytest.param(403, PermissionDenied),
        pytest.param(404, UserNotRegistered),
    ],
)
def test_authentication_errors(demo_api, SMARTID_DEMO_ID_CODE_EE, status, exc):
    api_url = demo_api.Actions.AUTH.format(id_code=SMARTID_DEMO_ID_CODE_EE, country=Countries.ESTONIA)
    with requests_mock.mock() as m:
        m.post(demo_api.api_url(api_url), status_code=status)

        with pytest.raises(exc):
            demo_api.authenticate(SMARTID_DEMO_ID_CODE_EE, Countries.ESTONIA)


def test_authentication_invalid_id_code(demo_api):
    with pytest.raises(InvalidIdCode):
        demo_api.authenticate("0", Countries.ESTONIA)


def run_authentication_flow(demo_api: SmartIDService, id_number, country, hash_type=HASH_SHA512, timeout=30):
    """Run full authentication flow w/ a hash algorithm

    id_number from https://github.com/SK-EID/smart-id-documentation/wiki/Environment-technical-parameters

    :param demo_api:
    :param str id_number:
    :param str country:
    :param str hash_type:
    :param int timeout:
    :rtype: AuthenticateStatusResult
    """
    res = demo_api.authenticate(id_number, country, hash_type=hash_type)
    assert isinstance(res, AuthenticateResult)

    # All fields must be set
    assert res.session_id
    assert res.hash_type == hash_type
    assert res.hash_value
    assert res.verification_code

    status_res = None  # type: AuthenticateStatusResult

    # Pull status (using a loop here since the remote might be slow)
    # Default timeout is 30s compared to Smart ID ~5m.
    # There is really no reason that it should take longer IF it is working
    end_time = time() + timeout
    while time() < end_time:
        try:
            status_res = demo_api.status(res.session_id, res.hash_value)
            break
        except ActionInProgress:
            sleep(1.0)

    if status_res is None:
        raise TimeoutError(f"Failed to get status in {timeout}s")

    assert isinstance(status_res, AuthenticateStatusResult)
    return status_res


@pytest.mark.slow
@pytest.mark.parametrize("hash_type", HASH_ALGORITHMS)
def test_all_hash_algorithms(demo_api, hash_type, SMARTID_DEMO_ID_CODE_EE):
    """Test full authentication flow w/ all hash algorithms

    Note: This does a real connection to Smart-ID api so it's marked as a slow test
    """
    status_res = run_authentication_flow(demo_api, SMARTID_DEMO_ID_CODE_EE, Countries.ESTONIA, hash_type)

    assert status_res.document_number == "PNOEE-10101010005-Z1B2-Q"
    assert status_res.certificate
    assert status_res.certificate_level == CERTIFICATE_LEVEL_QUALIFIED


@pytest.mark.slow
def test_authentication_flow_ee(demo_api, SMARTID_DEMO_ID_CODE_EE):
    """Test full authentication flow for EE

    Note: This does a real connection to Smart-ID api so it's marked as a slow test
    """
    status_res = run_authentication_flow(demo_api, SMARTID_DEMO_ID_CODE_EE, Countries.ESTONIA)

    assert status_res.document_number == f"PNOEE-{SMARTID_DEMO_ID_CODE_EE}-Z1B2-Q"
    assert status_res.certificate
    assert status_res.certificate_level == CERTIFICATE_LEVEL_QUALIFIED


@pytest.mark.slow
def test_authentication_flow_lv(demo_api, SMARTID_DEMO_ID_CODE_LV):
    """Test full authentication flow for LV

    Note: This does a real connection to Smart-ID api so it's marked as a slow test
    """
    status_res = run_authentication_flow(demo_api, SMARTID_DEMO_ID_CODE_LV, Countries.LATVIA)

    assert status_res.document_number == f"PNOLV-{SMARTID_DEMO_ID_CODE_LV}-SGT7-Q"
    assert status_res.certificate
    assert status_res.certificate_level == CERTIFICATE_LEVEL_QUALIFIED


@pytest.mark.slow
def test_authentication_flow_lt(demo_api, SMARTID_DEMO_ID_CODE_LT):
    """Test full authentication flow for LT

    Note: This does a real connection to Smart-ID api so it's marked as a slow test
    """
    status_res = run_authentication_flow(demo_api, SMARTID_DEMO_ID_CODE_LT, Countries.LITHUANIA)

    assert status_res.document_number == f"PNOLT-{SMARTID_DEMO_ID_CODE_LT}-Z52N-Q"
    assert status_res.certificate
    assert status_res.certificate_level == CERTIFICATE_LEVEL_QUALIFIED


def run_sign_flow(demo_api: SmartIDService, id_number=None, country=None, doc_num=None, signed_data=None, timeout=60):
    """Run full sign flow w/ a hash algorithm

    id_number from https://github.com/SK-EID/smart-id-documentation/wiki/Environment-technical-parameters

    :param demo_api:
    :param str id_number:
    :param str country:
    :param str doc_num: alternatively to id_number+country
    :param bytes signed_data: content to sign
    :param int timeout: how long to wait for a result. NOTE: if user input is expected, ensure at least 1 minute
    :rtype: SignStatusResult
    """
    if signed_data is None:
        signed_data = b"Test"
    if doc_num is not None:
        res = demo_api.sign_by_document_number(doc_num, signed_data)
    else:
        res = demo_api.sign(id_number, country, signed_data)
    assert isinstance(res, SignResult)

    # All fields must be set
    assert res.session_id

    status_res = None  # type: SignStatusResult

    # Poll status (using a loop here since the remote might be slow)
    # We timeout in 15s compared to Smart ID ~5m.
    # There is really no reason that it should take longer IF it is working
    end_time = time() + timeout
    while status_res is None and time() < end_time:
        try:
            status_res = demo_api.sign_status(res.session_id, generate_hash(HASH_SHA256, signed_data))
        except ActionInProgress:
            sleep(1.0)

    assert isinstance(status_res, SignStatusResult)
    return status_res


@pytest.mark.slow
def test_sign_flow_ee(demo_api, SMARTID_DEMO_ID_CODE_EE):
    """Test full sign flow for EE

    Note: This does a number of real connections to Smart-ID apis so it's marked as a slow test
    """
    file_name = "test.txt"
    data = b"Hello World!"
    mime_type = "text/plain"

    # Authenticate user
    auth_result = run_authentication_flow(demo_api, SMARTID_DEMO_ID_CODE_EE, Countries.ESTONIA)

    # Select user's certificate
    subject_cert, _ = demo_api.select_signing_certificate(document_number=auth_result.document_number)

    # Generate a XAdES signature
    xs: XmlSignature = (
        XmlSignature.create()
        .add_document(file_name, data, mime_type)
        .set_certificate(subject_cert)
        .update_signed_info()
    )

    # Sign the XAdES structure
    sign_result = run_sign_flow(demo_api, doc_num=auth_result.document_number, signed_data=xs.signed_data())
    xs.set_signature_value(sign_result.signature)

    # prove that all went right
    xs.verify()

    issuer_cn = xs.get_certificate_issuer_common_name()
    issuer_cert = esteid_certificates.get_certificate(issuer_cn)

    # Get an OCSP status confirmation
    ocsp = OCSP(url=OCSP_DEMO_URL)
    ocsp.validate(subject_cert, issuer_cert, sign_result.signature)

    # Embed the OCSP response
    xs.set_ocsp_response(ocsp)

    # Get a signature TimeStamp
    tsa = TSA(url=TSA_DEMO_URL)
    tsr = tsa.get_timestamp(xs.get_timestamped_message())
    xs.set_timestamp_response(tsr)

    # Write results to a bdoc file
    bdoc_file = Container()
    bdoc_file.add_file(file_name, data, mime_type)
    bdoc_file.add_signature(xs)

    bdoc_file.verify_container()
    bdoc_file.verify_signatures()



from typing import List

from django.http import HttpRequest, JsonResponse
from django.views import View

from .actions import BaseAction
from .types import DataFile


class GenericDigitalSignViewMixin(object):
    request: HttpRequest

    def get_files(self) -> List[DataFile]:
        """
        This should be implemented on view level, and should return a list of files that should be digitally signed.

        Will be ignored if `get_bdoc_container_file()` returns not None
        """
        return []

    def get_bdoc_container_file(self) -> str:
        """
        Returns path to a container that should be digitally signed.

        If there is a container, then get_files() will be ignored, and a new signature will be added to the container.
        """
        pass


class ApiView(View):
    def dispatch(self, request, *args, **kwargs):
        response = super().dispatch(request, *args, **kwargs)
        if isinstance(response, (dict, list, tuple)):
            return JsonResponse(response)

        return response


class SignStatusViewMixin:
    """Gets the status of signing and finalizes the process if necessary."""

    ACTION_CLASS: BaseAction

    def build_action_kwargs(self, request):
        return {}

    def post(self, request, *args, **kwargs):
        action_kwargs = self.build_action_kwargs(request)
        return self.ACTION_CLASS.do_action(self, **action_kwargs)

    def get(self, request, *args, **kwargs):
        return self.post(request, *args, **kwargs)



import pytest

import requests
import requests_mock
from django.core.exceptions import ImproperlyConfigured
from requests.exceptions import ConnectTimeout

from ...constants import SMART_ID_DEMO_URL, SMART_ID_LIVE_URL
from ...exceptions import (
    BadRequest,
    EsteidError,
    InvalidCredentials,
    OfflineError,
    UnsupportedClientImplementation,
    UpstreamServiceError,
)
from .. import SmartIDError
from ..base import SmartIDService
from ..i18n import TranslatedSmartIDService


def test_error():
    assert SmartIDError is EsteidError


@pytest.mark.parametrize("exc", [ConnectionError, requests.ConnectionError, ConnectTimeout])
def test_smartid_invoke_timeout(demo_api, exc):
    with requests_mock.mock() as m:
        m.get(demo_api.api_url(""), exc=exc)

        with pytest.raises(OfflineError):
            demo_api.invoke("")


@pytest.mark.parametrize(
    "status_code,exc",
    [
        (401, InvalidCredentials),
        (480, UnsupportedClientImplementation),
        (580, OfflineError),
        (502, OfflineError),
        (503, OfflineError),
        (504, OfflineError),
        (400, BadRequest),
        (500, UpstreamServiceError),
    ],
)
def test_smartid_invoke_errors(demo_api, status_code, exc):
    with requests_mock.mock() as m:
        m.get(demo_api.api_url(""), status_code=status_code)

        with pytest.raises(exc):
            demo_api.invoke("")


def test_smartid_service(demo_api):
    assert demo_api.api_root == SMART_ID_DEMO_URL

    service = SmartIDService("00000000-0000-0000-0000-000000000000", "test", SMART_ID_LIVE_URL)
    assert service.api_root == SMART_ID_LIVE_URL
    assert service.rp_uuid == "00000000-0000-0000-0000-000000000000"
    assert service.rp_name == "test"


def test_smartid_translated_service(i18n_demo_api):
    assert i18n_demo_api.api_root == SMART_ID_DEMO_URL


def test_smartid_translated_service_test_mode_off(override_esteid_settings):
    with override_esteid_settings(
        SMART_ID_TEST_MODE=False,
        SMART_ID_SERVICE_UUID="00000000-0000-0000-0000-000000000000",
        SMART_ID_SERVICE_NAME="test",
    ):
        service = TranslatedSmartIDService.get_instance()
        assert service.api_root == SMART_ID_LIVE_URL
        assert service.rp_uuid == "00000000-0000-0000-0000-000000000000"
        assert service.rp_name == "test"


def test_smartid_translated_service_requires_creds_for_live(override_esteid_settings):
    with override_esteid_settings(SMART_ID_TEST_MODE=False):
        with pytest.raises(ImproperlyConfigured, match="SMART_ID_SERVICE_NAME and SMART_ID_SERVICE_UUID"):
            TranslatedSmartIDService.get_instance()

    with override_esteid_settings(
        SMART_ID_TEST_MODE=False,
        SMART_ID_SERVICE_UUID="00000000-0000-0000-0000-000000000000",
    ):
        with pytest.raises(ImproperlyConfigured, match="SMART_ID_SERVICE_NAME and SMART_ID_SERVICE_UUID"):
            TranslatedSmartIDService.get_instance()

    with override_esteid_settings(SMART_ID_TEST_MODE=False, SMART_ID_SERVICE_NAME="name"):
        with pytest.raises(ImproperlyConfigured, match="SMART_ID_SERVICE_NAME and SMART_ID_SERVICE_UUID"):
            TranslatedSmartIDService.get_instance()



import logging
import re
from typing import List, Optional

from pyasice import Container, XmlSignature

from esteid.exceptions import ActionInProgress, InvalidIdCode, InvalidParameter, InvalidParameters
from esteid.signing import DataFile, Signer
from esteid.signing.types import InterimSessionData, PredictableDict

from ..util import id_code_ee_is_valid
from .constants import Languages
from .i18n import TranslatedMobileIDService


logger = logging.getLogger(__name__)


PHONE_NUMBER_REGEXP = re.compile(r"^\+37[02]\d{7,8}$")  # Mobile ID supports Estonian and Lithuanian phones


class UserInput(PredictableDict):
    phone_number: str
    id_code: str
    language: Optional[str]

    def is_valid(self, raise_exception=True):
        result = super().is_valid(raise_exception=raise_exception)
        if result:
            if not (self.phone_number and re.match(PHONE_NUMBER_REGEXP, self.phone_number)):
                raise InvalidParameter(param="phone_number")
            if not id_code_ee_is_valid(self.id_code):
                raise InvalidIdCode
            if not (self.get("language") and self.language in Languages.ALL):
                self.language = Languages.ENG
        return result


class MobileIdSessionData(InterimSessionData):
    session_id: str


class MobileIdSigner(Signer):
    phone_number: str
    id_code: str
    language: str

    SessionData = MobileIdSessionData
    session_data: MobileIdSessionData

    def setup(self, initial_data: dict = None):
        """
        Receives user input via POST: `id_code`, `phone_number`, `language`
        """
        if not isinstance(initial_data, dict):
            raise InvalidParameters("Missing required parameters")

        user_input = UserInput(**initial_data)

        try:
            user_input.is_valid()
        except (InvalidIdCode, InvalidParameter):
            # Just to be explicit
            raise
        except ValueError as e:
            raise InvalidParameters("Invalid parameters") from e

        self.id_code = user_input.id_code
        self.phone_number = user_input.phone_number
        self.language = user_input.language

    def prepare(self, container_file=None, files: List[DataFile] = None) -> dict:
        container = self.open_container(container_file, files)

        service = TranslatedMobileIDService.get_instance()

        certificate = service.get_certificate(self.id_code, self.phone_number)

        xml_sig = container.prepare_signature(certificate)

        sign_session = service.sign(self.id_code, self.phone_number, xml_sig.signed_data(), language=self.language)

        self.save_session_data(
            digest=sign_session.digest,
            container=container,
            xml_sig=xml_sig,
            session_id=sign_session.session_id,
        )

        return {
            "verification_code": sign_session.verification_code,
        }

    def finalize(self, data: dict = None) -> Container:
        digest = self.session_data.digest
        session_id = self.session_data.session_id

        temp_signature_file = self.session_data.temp_signature_file
        temp_container_file = self.session_data.temp_container_file

        with open(temp_signature_file, "rb") as f:
            xml_sig = XmlSignature(f.read())

        service = TranslatedMobileIDService.get_instance()

        try:
            status = service.sign_status(session_id, xml_sig.get_certificate_value(), digest)
        except ActionInProgress:
            # Just to be explicit about the InProgress status
            raise

        xml_sig.set_signature_value(status.signature)

        self.finalize_xml_signature(xml_sig)

        container = Container.open(temp_container_file)
        container.add_signature(xml_sig)

        return container

    def save_session_data(self, *, digest: bytes, container: Container, xml_sig: XmlSignature, session_id: str):
        data_obj = self.session_data
        data_obj.session_id = session_id

        super().save_session_data(digest=digest, container=container, xml_sig=xml_sig)



from collections import namedtuple


AuthenticateResult = namedtuple(
    "AuthenticateResult",
    [
        "session_id",
        "hash_type",
        "digest",
        "verification_code",
    ],
)

AuthenticateStatusResult = namedtuple(
    "AuthenticateStatusResult",
    [
        "signature",
        "signature_algorithm",
        "certificate",  # DER-encoded certificate
    ],
)

SignResult = namedtuple(
    "SignResult",
    [
        "session_id",
        "digest",
        "verification_code",
    ],
)

# Note: MobileID doesn't return a certificate for SignStatus. It is set from a previous call to `/certificate`
SignStatusResult = namedtuple(
    "SignStatusResult",
    [
        "signature",
        "signature_algorithm",
        "certificate",
    ],
)



import hashlib
from time import sleep, time
from unittest.mock import patch

import pytest

import requests_mock

import pyasice

from ...constants import HASH_ALGORITHMS, HASH_SHA256, HASH_SHA384, HASH_SHA512
from ...exceptions import (
    ActionInProgress,
    BadRequest,
    CanceledByUser,
    SessionDoesNotExist,
    SignatureVerificationError,
    UserNotRegistered,
    UserTimeout,
)
from ...util import generate_hash
from .. import MobileIDError
from ..base import MobileIDService
from ..constants import EndResults
from ..types import AuthenticateResult, AuthenticateStatusResult


def run_authentication_flow(demo_mid_api, id_number, phone_number, hash_type=HASH_SHA256):
    """Run full authentication flow w/ a hash algorithm

    test data from:
    https://github.com/SK-EID/MID/wiki/Test-number-for-automated-testing-in-DEMO

    :param MobileIDService demo_mid_api:
    :param str id_number:
    :param str phone_number:
    :param str hash_type:
    :rtype: AuthenticateStatusResult
    """
    res = demo_mid_api.authenticate(id_code=id_number, phone_number=phone_number, hash_type=hash_type)
    assert isinstance(
        res, AuthenticateResult
    ), "Expected `.authenticate()` to return AuthenticateResult, got {}".format(res)

    # All fields must be set
    assert res.session_id
    assert res.hash_type == hash_type
    assert res.digest

    status_res = None  # type: AuthenticateStatusResult

    # Pull status (using a loop here since the remote might be slow)
    # We timeout in 15s compared to Smart ID ~5m.
    # There is really no reason that it should take longer IF it is working
    end_time = time() + 15
    while status_res is None and time() < end_time:
        try:
            status_res = demo_mid_api.status(res.session_id, res.digest)
        except ActionInProgress:
            sleep(1.0)

    assert isinstance(status_res, AuthenticateStatusResult)
    return status_res


@pytest.mark.parametrize("hash_type", HASH_ALGORITHMS)
def test_mobileid_authentication(demo_mid_api, hash_type, MID_DEMO_PIN_EE_OK, MID_DEMO_PHONE_EE_OK):
    raw_data = b"Hello World!"
    response_data = {"sessionID": "FAKE"}
    verification_codes = {
        HASH_SHA256: "4073",
        HASH_SHA384: "6090",
        HASH_SHA512: "4296",
    }

    known_hashes = {
        HASH_SHA256: hashlib.sha256(raw_data).digest(),
        HASH_SHA384: hashlib.sha384(raw_data).digest(),
        HASH_SHA512: hashlib.sha512(raw_data).digest(),
    }

    with patch("esteid.mobileid.base.secure_random", return_value=raw_data):
        with patch("esteid.mobileid.base.generate_hash", return_value=known_hashes[hash_type]) as mock:
            with patch.object(demo_mid_api, "invoke", return_value=response_data):
                res = demo_mid_api.authenticate(MID_DEMO_PIN_EE_OK, MID_DEMO_PHONE_EE_OK, hash_type=hash_type)

                mock.assert_called_with(hash_type, raw_data)

                assert isinstance(res, AuthenticateResult)
                assert res.session_id == "FAKE"
                assert res.hash_type == hash_type
                assert res.digest == generate_hash(hash_type, raw_data)
                assert res.verification_code == verification_codes[hash_type]


def test_mobileid_authentication_400(demo_mid_api, MID_DEMO_PHONE_EE_OK, MID_DEMO_PIN_EE_OK):
    with requests_mock.mock() as m:
        m.post(demo_mid_api.api_url(demo_mid_api.Actions.AUTH), status_code=400)
        with pytest.raises(BadRequest):
            demo_mid_api.authenticate(MID_DEMO_PIN_EE_OK, MID_DEMO_PHONE_EE_OK)


def test_mobileid_status(demo_mid_api, static_certificate, mid_auth_result, mid_auth_status_response):
    with patch.object(demo_mid_api, "invoke", return_value=mid_auth_status_response):
        res = demo_mid_api.status(mid_auth_result.session_id, mid_auth_result.digest)

        assert isinstance(res, AuthenticateStatusResult)

        assert res.certificate == static_certificate


def test_mobileid_status_signature_verification_error(demo_mid_api, mid_auth_status_response, mid_auth_result):
    with patch.object(demo_mid_api, "invoke", return_value=mid_auth_status_response):
        with patch.object(pyasice, "verify") as mock_verify:
            mock_verify.side_effect = pyasice.SignatureVerificationError
            with pytest.raises(SignatureVerificationError):
                demo_mid_api.status(mid_auth_result.session_id, mid_auth_result.digest)


def test_mobileid_status_state_running(demo_mid_api):
    response_data = {
        "state": MobileIDService.ProcessingStates.RUNNING,
    }

    with patch.object(demo_mid_api, "invoke", return_value=response_data):
        with pytest.raises(ActionInProgress) as exc_info:
            demo_mid_api.status(session_id="FAKE", digest=b"")

        # session_id should be in the message
        assert "FAKE" in str(exc_info.value)


@pytest.mark.parametrize(
    "end_result_code,exc",
    [
        (EndResults.TIMEOUT, UserTimeout),
        (EndResults.USER_CANCELLED, CanceledByUser),
        (EndResults.NOT_MID_CLIENT, UserNotRegistered),
        ("$unknown$", MobileIDError),
    ],
)
def test_mobileid_status_end_result(demo_mid_api, end_result_code, exc):
    response_data = {
        "state": MobileIDService.ProcessingStates.COMPLETE,
        "result": end_result_code,
    }

    with patch.object(demo_mid_api, "invoke", return_value=response_data):
        with pytest.raises(exc):
            demo_mid_api.status(session_id="FAKE", digest=b"")


def test_mobileid_status_unexpected_end_result(demo_mid_api):
    response_data = {
        "state": MobileIDService.ProcessingStates.COMPLETE,
        "result": "$RESULT$",
    }

    with patch.object(demo_mid_api, "invoke", return_value=response_data):
        with pytest.raises(MobileIDError) as exc_info:
            demo_mid_api.status(session_id="FAKE", digest=b"")

        assert "$RESULT$" in str(exc_info.value)


@pytest.mark.parametrize(
    "status,exc",
    [
        pytest.param(400, BadRequest),
        pytest.param(404, SessionDoesNotExist),
    ],
)
def test_mobileid_status_errors(demo_mid_api, status, exc):
    session_status_url = demo_mid_api.Actions.SESSION_STATUS.format(action=demo_mid_api.Actions.AUTH, session_id="FAKE")
    with requests_mock.mock() as m:
        m.get(demo_mid_api.api_url(session_status_url), exc=exc)
        with pytest.raises(exc):
            demo_mid_api.status(session_id="FAKE", digest=b"")


@pytest.mark.slow
@pytest.mark.parametrize("hash_type", [*HASH_ALGORITHMS, None])
def test_mobileid_authentication_flow_ee(demo_mid_api, hash_type, MID_DEMO_PIN_EE_OK, MID_DEMO_PHONE_EE_OK):
    """Test full authentication flow for EE

    Note: This does a real connection to Mobile-ID api so it's marked as a slow test
    """
    kwargs = {} if hash_type is None else dict(hash_type=hash_type)
    status_res = run_authentication_flow(
        demo_mid_api, id_number=MID_DEMO_PIN_EE_OK, phone_number=MID_DEMO_PHONE_EE_OK, **kwargs
    )

    assert status_res.certificate



import binascii
import logging
from typing import List

from oscrypto.asymmetric import Certificate as OsCryptoCertificate
from oscrypto.asymmetric import load_certificate

import pyasice

from esteid.exceptions import InvalidParameter, SignatureVerificationError
from esteid.signing import DataFile, Signer
from esteid.types import Signer as SignerInfo


logger = logging.getLogger(__name__)


class IdCardSigner(Signer):
    certificate: bytes

    _certificate_handle: "OsCryptoCertificate"

    @property
    def id_code(self) -> str:
        try:
            certificate_handle = self._certificate_handle
        except AttributeError as e:
            raise AttributeError("Attribute id_code not available: certificate not provided") from e

        signer_info = SignerInfo.from_certificate(certificate_handle)
        return signer_info.id_code

    def setup(self, initial_data: dict = None):
        """
        Receives a user certificate from the front end
        """
        try:
            certificate_hex = initial_data["certificate"]
        except (TypeError, KeyError) as e:
            raise InvalidParameter("Missing required parameter 'certificate'", param="certificate") from e

        try:
            certificate = binascii.a2b_hex(certificate_hex)
        except (TypeError, ValueError) as e:
            raise InvalidParameter(
                "Failed to decode parameter `certificate` from hex-encoding", param="certificate"
            ) from e

        try:
            self._certificate_handle = load_certificate(certificate)
        except ValueError as e:
            raise InvalidParameter(
                "Failed to recognize `certificate` as a supported certificate format", param="certificate"
            ) from e

        self.certificate = certificate

    def prepare(self, container_file=None, files: List[DataFile] = None) -> dict:
        container = self.open_container(container_file, files)
        xml_sig = container.prepare_signature(self.certificate)

        # Note: uses default digest algorithm (sha256)
        signed_digest = xml_sig.digest()

        self.save_session_data(digest=signed_digest, container=container, xml_sig=xml_sig)

        return {
            # hex-encoded digest to be consumed by the hwcrypto.js library
            "digest": binascii.b2a_hex(signed_digest).decode(),
        }

    def finalize(self, data: dict = None) -> pyasice.Container:
        try:
            signature_value = data["signature_value"]
        except (TypeError, KeyError) as e:
            raise InvalidParameter("Missing required parameter 'signature_value'", param="signature_value") from e

        try:
            signature_value = binascii.a2b_hex(signature_value)
        except (TypeError, ValueError) as e:
            raise InvalidParameter(
                "Failed to decode parameter `signature_value` from hex-encoding", param="signature_value"
            ) from e

        temp_signature_file = self.session_data.temp_signature_file
        temp_container_file = self.session_data.temp_container_file
        digest = self.session_data.digest

        with open(temp_signature_file, "rb") as f:
            xml_sig = pyasice.XmlSignature(f.read())

        try:
            pyasice.verify(xml_sig.get_certificate_value(), signature_value, digest, prehashed=True)
        except pyasice.SignatureVerificationError as e:
            raise SignatureVerificationError from e

        container = pyasice.Container.open(temp_container_file)

        xml_sig.set_signature_value(signature_value)

        self.finalize_xml_signature(xml_sig)

        container.add_signature(xml_sig)

        return container



XXXXXXXXXX
XXXXXXX
XXXXXXXXXXX
XXXXXXXXX
XXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXX
XXXXXXXXX
XXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXX
XXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXX
XXXXXXXXXXXXXX
XXXXXXXXXXXXXX
XXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXX


"""
Compatibility utilities that provide signed container info
in the same format as the DDS service did.
"""
import base64
import hashlib

import attr

from pyasice import Container, XmlSignature

from esteid.types import ResponderCertificate, SignedDocInfo, Signer


def signature_info(xml_signature: XmlSignature):
    subject_cert = xml_signature.get_certificate()
    ocsp_cert = xml_signature.get_ocsp_response().get_responder_certs()[0]
    signing_time = xml_signature.get_signing_time()

    return {
        "id": xml_signature._get_node("ds:Signature").attrib["Id"],
        "signing_time": signing_time,
        "status": True,
        "signature_production_place": None,
        "signer": attr.asdict(Signer.from_certificate(subject_cert)),
        "confirmation": {
            # if ever needed
            "produced_at": signing_time,
            "responder_id": "OCSP",
            "responder_certificate": ResponderCertificate.from_certificate(ocsp_cert),
        },
        "signer_role": [{"certified": 0, "role": None}],
    }


def container_info(bdoc_container: Container, full=True):
    if full:
        data_files_info = [
            {
                "digest_type": "sha256",
                "digest_value": base64.b64encode(hashlib.sha256(content).digest()).decode(),
                "filename": name,
                "id": name,
                "mime_type": mimetype,
                "content_type": mimetype,
                "size": len(content),
            }
            for name, content, mimetype in bdoc_container.iter_data_files()
        ]
    else:
        data_files_info = [
            {
                "filename": name,
            }
            for name in bdoc_container.data_file_names
        ]
    return SignedDocInfo.from_dict(
        {
            "data_file_info": data_files_info,
            "signature_info": [signature_info(signature) for signature in bdoc_container.iter_signatures()],
        }
    )



from unittest.mock import patch

import pytest

import requests_mock

import pyasice

from esteid.exceptions import (
    ActionInProgress,
    BadRequest,
    CanceledByUser,
    EsteidError,
    SessionDoesNotExist,
    SignatureVerificationError,
    UpstreamServiceError,
    UserTimeout,
)
from esteid.smartid.base import SmartIDService
from esteid.smartid.constants import CERTIFICATE_LEVEL_QUALIFIED, EndResults
from esteid.smartid.types import AuthenticateStatusResult


def test_status(demo_api, static_certificate, static_auth_result, static_status_response):
    with patch.object(demo_api, "invoke", return_value=static_status_response):
        res = demo_api.status(static_auth_result.session_id, static_auth_result.hash_value)

        assert isinstance(res, AuthenticateStatusResult)

        assert res.document_number == "$documentNumber$"
        assert res.certificate == static_certificate
        assert res.certificate_level == CERTIFICATE_LEVEL_QUALIFIED


def test_status_signature_verification(demo_api, static_auth_result, static_status_response):
    with patch.object(demo_api, "invoke", return_value=static_status_response):
        with patch.object(pyasice, "verify", side_effect=pyasice.SignatureVerificationError):
            with pytest.raises(SignatureVerificationError):
                demo_api.status(static_auth_result.session_id, static_auth_result.hash_value)


def test_status_state_running(demo_api):
    response_data = {
        "state": SmartIDService.ProcessingStates.RUNNING,
    }

    with patch.object(demo_api, "invoke", return_value=response_data):
        with pytest.raises(ActionInProgress) as exc_info:
            demo_api.status(session_id="FAKE", hash_value=b"")

        # session_id should be in the message
        assert "FAKE" in str(exc_info.value)


@pytest.mark.parametrize(
    "end_result_code,exc",
    [
        (EndResults.USER_REFUSED, CanceledByUser),
        (EndResults.TIMEOUT, UserTimeout),
        (EndResults.DOCUMENT_UNUSABLE, UpstreamServiceError),
        ("$unknown$", EsteidError),
    ],
)
def test_status_end_result(demo_api, end_result_code, exc):
    response_data = {
        "state": SmartIDService.ProcessingStates.COMPLETE,
        "result": {
            "endResult": end_result_code,
        },
    }

    with patch.object(demo_api, "invoke", return_value=response_data):
        with pytest.raises(exc):
            demo_api.status(session_id="FAKE", hash_value=b"")


def test_status_unexpected_end_result(demo_api):
    response_data = {
        "state": SmartIDService.ProcessingStates.COMPLETE,
        "result": {
            "endResult": "$RESULT$",
        },
    }

    with patch.object(demo_api, "invoke", return_value=response_data):
        with pytest.raises(EsteidError) as exc_info:
            demo_api.status(session_id="FAKE", hash_value=b"")

        assert "$RESULT$" in str(exc_info.value)


@pytest.mark.parametrize(
    "status,exc",
    [
        pytest.param(400, BadRequest),
        pytest.param(404, SessionDoesNotExist),
    ],
)
def test_status_400(demo_api, status, exc):
    with requests_mock.mock() as m:
        m.get(demo_api.api_url(demo_api.Actions.SESSION_STATUS.format(session_id="FAKE")), status_code=status)

        with pytest.raises(exc):
            demo_api.status(session_id="FAKE", hash_value=b"")



from esteid import settings


def esteid_services(*_, **__):
    return {
        "ESTEID_DEMO": settings.ESTEID_DEMO,
        "ID_CARD_ENABLED": settings.ID_CARD_ENABLED,
        "MOBILE_ID_ENABLED": settings.MOBILE_ID_ENABLED,
        "MOBILE_ID_TEST_MODE": settings.MOBILE_ID_TEST_MODE,
        "SMART_ID_ENABLED": settings.SMART_ID_ENABLED,
        "SMART_ID_TEST_MODE": settings.SMART_ID_TEST_MODE,
    }



import json
from http import HTTPStatus
from unittest.mock import Mock, patch

import pytest

from esteid.exceptions import (
    ActionInProgress,
    AlreadySignedByUser,
    CanceledByUser,
    EsteidError,
    InvalidParameters,
    OfflineError,
    UpstreamServiceError,
    UserNotRegistered,
)
from esteid.types import Signer as SignerData

from ..views import SignViewMixin


@pytest.fixture()
def signatory_id_code(signed_container):
    signature = next(signed_container.iter_signatures())
    subject_cert = signature.get_certificate()
    signatory = SignerData.from_certificate(subject_cert)
    return signatory.id_code


@pytest.fixture()
def signer_class():
    signer_class = Mock(name="signer class")
    signer_class.start_session.return_value = signer_class()
    signer_class.load_session.return_value = signer_class()
    return signer_class


@pytest.fixture()
def signing_view(signer_class):
    the_view = SignViewMixin()
    with patch.object(the_view, "select_signer_class", return_value=signer_class), patch.object(
        the_view, "get_container"
    ):
        yield the_view


@pytest.mark.parametrize(
    "container,allow_sign_many,result",
    [
        pytest.param(None, True, None, id="No container, allowed - OK"),
        pytest.param(None, False, None, id="No container, not allowed - OK"),
        pytest.param(True, True, None, id="Container, allowed - OK"),
        pytest.param(True, False, AlreadySignedByUser, id="Container, not allowed - ERROR"),
    ],
)
def test_signing_views_check_eligibility(
    signatory_id_code, container, allow_sign_many, result, signed_container, override_esteid_settings
):
    if container:
        container = signed_container

    signer = Mock(id_code=signatory_id_code)
    signer_not_signatory = Mock(id_code="12345678912")
    view = SignViewMixin()

    with override_esteid_settings(ESTEID_ALLOW_ONE_PARTY_SIGN_TWICE=allow_sign_many):
        # Signing by a different party should always succeed
        view.check_eligibility(signer_not_signatory, container)

        if result is AlreadySignedByUser:
            with pytest.raises(AlreadySignedByUser):
                view.check_eligibility(signer, container)
        else:
            view.check_eligibility(signer, container)


def test_signing_views_canceled_by_user(signing_view, signer_class):
    signer_class().finalize.side_effect = CanceledByUser

    with patch.object(signing_view, "handle_user_cancel"):
        result = signing_view.patch(Mock())

        signing_view.handle_user_cancel.assert_called_once()

        assert result.status_code == CanceledByUser.status
        assert json.loads(result.content) == {
            "status": "error",
            "error": "CanceledByUser",
            "message": str(CanceledByUser.default_message),
        }


@pytest.mark.parametrize(
    "error",
    [
        OfflineError(service="service"),
        UpstreamServiceError(service="service"),
    ],
)
def test_signing_views_esteid_error_handling_finalize(error: EsteidError, signing_view, signer_class):
    signer_class().finalize.side_effect = error

    result = signing_view.patch(Mock())

    assert result.status_code == error.status
    assert json.loads(result.content) == {
        "status": "error",
        "error": error.__class__.__name__,
        "message": str(error.default_message).format(service="service"),
    }


def test_signing_views_exception_handling_finalize(signing_view, signer_class):
    signer_class().finalize.side_effect = Exception("unhandled")

    result = signing_view.patch(Mock())

    assert result.status_code == HTTPStatus.INTERNAL_SERVER_ERROR
    assert json.loads(result.content) == {
        "status": "error",
        "error": "Internal error",
        "message": "Internal server error",
    }


def test_signing_views_action_in_progress(signing_view, signer_class):
    progress_data = {"verification_code": 1111}
    signer_class().finalize.side_effect = ActionInProgress("In progress", data=progress_data)

    result = signing_view.patch(Mock())

    assert result.status_code == ActionInProgress.status == HTTPStatus.ACCEPTED
    assert json.loads(result.content) == {"status": "pending", **progress_data}


@pytest.mark.parametrize(
    "error",
    [
        InvalidParameters(),
        OfflineError(service="service"),
        UserNotRegistered(),
        UpstreamServiceError(service="service"),
    ],
)
def test_signing_views_esteid_error_handling_prepare(error: EsteidError, signing_view, signer_class):
    signer_class().prepare.side_effect = error

    result = signing_view.post(Mock())

    assert result.status_code == error.status
    assert json.loads(result.content) == {
        "status": "error",
        "error": error.__class__.__name__,
        "message": str(error.default_message).format(service="service"),  # .format(): unused arguments are ignored
    }


def test_signing_views_exception_handling_prepare(signing_view, signer_class):
    signer_class().prepare.side_effect = Exception("unhandled")

    result = signing_view.post(Mock())

    assert result.status_code == HTTPStatus.INTERNAL_SERVER_ERROR
    assert json.loads(result.content) == {
        "status": "error",
        "error": "Internal error",
        "message": "Internal server error",
    }



import base64
import os
from tempfile import NamedTemporaryFile
from unittest.mock import patch

import pytest

from esteid.mobileid import MobileIdSigner
from esteid.mobileid import signer as signer_module

from ...exceptions import InvalidIdCode, InvalidParameter, InvalidParameters, SigningSessionDoesNotExist


@pytest.fixture()
def mobileidsigner():
    signer = MobileIdSigner({}, initial=True)
    with patch.object(signer, "open_container"), patch.object(signer, "save_session_data"):
        yield signer


@pytest.fixture()
def mobileid_session_data():
    with NamedTemporaryFile("wb", delete=False) as f:
        f.write(b"xml signature data")
    yield {
        "digest_b64": "MTIz",
        "temp_signature_file": f.name,
        "temp_container_file": "...",
        "session_id": "...",
        "timestamp": 1,
    }
    os.remove(f.name)


@pytest.fixture()
def mobileidservice():
    with patch.object(signer_module, "TranslatedMobileIDService") as service_cls:
        yield service_cls


@pytest.mark.parametrize(
    "data,error",
    [
        pytest.param(
            None,
            InvalidParameters,
            id="No data",
        ),
        pytest.param(
            {},
            InvalidParameters,
            id="Empty data",
        ),
        pytest.param(
            {"id_code": "asdf", "phone_number": "asdf"},
            InvalidParameter,
            id="Invalid Phone",
        ),
        pytest.param(
            {"id_code": "asdf", "phone_number": "+37200000766"},
            InvalidIdCode,
            id="Invalid ID Code",
        ),
        pytest.param(
            {"id_code": "60001019906", "phone_number": "+37200000766"},
            None,
            id="ID Code and Phone OK",
        ),
        pytest.param(
            {"id_code": "60001019906", "phone_number": "+37200000766", "language": "EST"},
            None,
            id="ID Code and Phone OK",
        ),
    ],
)
def test_mobileidsigner_setup(data, error):
    signer = MobileIdSigner({}, initial=True)
    if error:
        with pytest.raises(error):
            signer.setup(data)
    else:
        signer.setup(data)
        assert signer.id_code == data["id_code"]
        assert signer.phone_number == data["phone_number"]
        assert signer.language == data.get("language", "ENG")


@pytest.mark.parametrize(
    "data,error",
    [
        pytest.param(None, SigningSessionDoesNotExist, id="No session data"),
        pytest.param({}, SigningSessionDoesNotExist, id="Empty session data"),
        pytest.param({"digest_b64": "asd"}, SigningSessionDoesNotExist, id="Broken session data"),
        pytest.param(
            {
                "digest_b64": "MTIz",
                "temp_signature_file": "a",
                "temp_container_file": "b",
                "session_id": "c",
                "timestamp": 1,
            },
            None,
            id="Good session data",
        ),
    ],
)
def test_mobileidsigner_load_session(data, error):
    if error:
        with pytest.raises(error):
            MobileIdSigner({MobileIdSigner._SESSION_KEY: data})
    else:
        signer = MobileIdSigner({MobileIdSigner._SESSION_KEY: data})
        assert signer.session_data.digest == b"123"
        assert signer.session_data.session_id == "c"


def test_mobileidsigner_prepare(mobileidsigner, MID_DEMO_PHONE_EE_OK, MID_DEMO_PIN_EE_OK, mobileidservice):
    mobileidsigner.setup({"id_code": MID_DEMO_PIN_EE_OK, "phone_number": MID_DEMO_PHONE_EE_OK})

    result = mobileidsigner.prepare(None, [])

    mobileidsigner.open_container.assert_called_once_with(None, [])
    container = mobileidsigner.open_container(...)

    mobileidservice.get_instance.assert_called_once()

    service = mobileidservice.get_instance()
    service.get_certificate.assert_called_once_with(MID_DEMO_PIN_EE_OK, MID_DEMO_PHONE_EE_OK)

    cert = service.get_certificate(...)

    container.prepare_signature.assert_called_once_with(cert)

    xml_sig = mobileidsigner.open_container().prepare_signature(...)
    xml_sig.signed_data.assert_called_once()

    service.sign.assert_called_once_with(
        MID_DEMO_PIN_EE_OK, MID_DEMO_PHONE_EE_OK, xml_sig.signed_data(), language=mobileidsigner.language
    )

    sign_session = service.sign(...)

    mobileidsigner.save_session_data.assert_called_once_with(
        digest=sign_session.digest,
        container=container,
        xml_sig=xml_sig,
        session_id=sign_session.session_id,
    )

    assert result["verification_code"] == sign_session.verification_code


@patch.object(signer_module, "Container")
@patch.object(signer_module, "XmlSignature")
def test_mobileidsigner_finalize(
    XmlSignatureMock, ContainerMock, MID_DEMO_PHONE_EE_OK, MID_DEMO_PIN_EE_OK, mobileidservice, mobileid_session_data
):
    mobileidsigner = MobileIdSigner({MobileIdSigner._SESSION_KEY: mobileid_session_data})
    with patch.object(mobileidsigner, "finalize_xml_signature"):
        result = mobileidsigner.finalize()

        XmlSignatureMock.assert_called_once_with(b"xml signature data")

        mobileidsigner.finalize_xml_signature.assert_called_once_with(XmlSignatureMock())

    xml_sig = XmlSignatureMock(...)

    service = mobileidservice.get_instance()
    service.sign_status.assert_called_once_with(
        mobileid_session_data["session_id"],
        xml_sig.get_certificate_value(),
        base64.b64decode(mobileid_session_data["digest_b64"]),
    )

    sign_status = service.sign_status(...)
    xml_sig.set_signature_value.assert_called_once_with(sign_status.signature)
    ContainerMock.open.assert_called_once_with(mobileid_session_data["temp_container_file"])
    container = ContainerMock.open(...)

    assert result == container



import base64
from tempfile import NamedTemporaryFile

from django.views.generic import DetailView
from rest_framework.views import APIView

from pyasice import Container

from esteid.signing import DataFile, SignViewDjangoMixin, SignViewRestMixin

# Register our signers
from ..idcard import IdCardSigner  # noqa
from ..mobileid import MobileIdSigner  # noqa
from ..smartid import SmartIdSigner  # noqa
from .signer import MyPostSigner, MySigner  # noqa


class BaseMethods:
    def get_files_to_sign(self, *args, **kwargs):
        files = self.request.session.get("__ddoc_files")
        if files:
            return [
                DataFile(
                    file_name,
                    content=base64.b64decode(file["content"]),
                    mime_type=file["content_type"],
                )
                for file_name, file in files.items()
            ]

        return []

    def save_container(self, container: Container, *args, **kwargs):
        with NamedTemporaryFile("wb", delete=False) as f:
            f.write(container.finalize().getbuffer())
        self.request.session["__ddoc_container_file"] = f.name


class SigningTestView(BaseMethods, SignViewDjangoMixin, DetailView):
    pass


class SigningTestRestView(BaseMethods, SignViewRestMixin, APIView):
    pass



import uuid

import pytest

from esteid.base_service import BaseSKService
from esteid.exceptions import EsteidError


def test_base_service_accepts_parameters():
    rp_uuid = uuid.uuid4()
    assert BaseSKService(rp_uuid, "name", "url").rp_uuid == str(rp_uuid)
    assert BaseSKService(str(rp_uuid), "name", "url").rp_uuid == str(rp_uuid)

    with pytest.raises(EsteidError, match="valid UUID"):
        BaseSKService("not uuid", "name", "url")



from http import HTTPStatus

from django.utils.translation import gettext_lazy as _


class EsteidError(Exception):
    """
    A generic Esteid error.

    Provides an interface for displaying a translated message to the user.

    Note: by default, any message passed as a first positional argument is ignored, because it only
    serves for better logging. The User Interface should display a verbose error message
    in case of a user-caused error so that user can take action accordingly. Errors caused by service malfunction
    are likely not informative to a user without malicious intent.

    If the default message contains {key} placeholders, the exception is supposed to be
    raised with kwargs, these will be passed on to <translated default message>.format(**kwargs).
    """

    status = HTTPStatus.INTERNAL_SERVER_ERROR

    default_message = _("Something went wrong")

    kwargs: dict

    def __init__(self, message=None, **kwargs):
        super().__init__(message)
        self.kwargs = kwargs

    def get_message(self):
        return str(self.default_message).format(**self.kwargs)

    def get_user_error(self):
        return {
            "error": self.__class__.__name__,  # error code that can be bound to JS translations
            # TODO review errors and add translations
            "message": self.get_message(),
        }


class ActionInProgress(EsteidError):
    """
    Raised when a request to the service is still in pending status,
    or when it is necessary to move on to the next step while passing some data on to the user
    (like after successful auth, we initiate a signing request and display verification code)
    """

    status = HTTPStatus.ACCEPTED

    default_message = _("Operation is in progress.")

    def __init__(self, message=None, data: dict = None):
        super().__init__(message)
        self.data = data or {}


# *** Misc errors resulting from misconfiguration or upstream service malfunction ***


class UpstreamServiceError(EsteidError):
    """A failure reported by upstream service which is not caused by user.

    Takes a `service` kwarg, e.g. Mobile ID / Smart ID
    """

    status = HTTPStatus.SERVICE_UNAVAILABLE

    default_message = _("The {service} service reported a failure.")

    def __init__(self, message=None, *, service, **kwargs):
        super().__init__(message)
        self.kwargs = {"service": service, **kwargs}


class OfflineError(UpstreamServiceError):
    default_message = _("The {service} service is unavailable at the moment. Please try again later.")


class PermissionDenied(UpstreamServiceError):
    """
    If a SmartID user has a certificate of level ADVANCED, it may be impossible to perform operations.
    """

    default_message = _("Operation with user certificate level ADVANCED not allowed.")

    def __init__(self, message=None, **kwargs):
        kwargs.setdefault("service", "smartid")
        super().__init__(message, **kwargs)


class UnsupportedClientImplementation(UpstreamServiceError):
    pass


# *** Errors that can not be determinably attributed to a service


class InvalidCredentials(EsteidError):
    pass


class BadRequest(EsteidError):
    pass


class SessionDoesNotExist(EsteidError):
    pass


class SignatureVerificationError(EsteidError):
    pass


# *** Signing Session errors ***


class SigningSessionError(EsteidError):
    """Base class for signing session errors"""

    status = HTTPStatus.CONFLICT


class SigningSessionExists(SigningSessionError):
    """A signing session already exists while not expected to exist (when initializing)"""

    default_message = _("A signing session is already in progress.")


class SigningSessionDoesNotExist(SigningSessionError):
    """A signing session does not exist while expected to exist"""

    default_message = _("No signing session found.")


# *** Errors caused by user action or inaction


class InvalidParameters(EsteidError):
    """Invalid initialization parameters received from the request."""

    status = HTTPStatus.BAD_REQUEST

    default_message = _("Invalid parameters.")


class InvalidParameter(InvalidParameters):
    """Invalid initialization parameter PARAM received from the request.

    Takes a `param` kwarg
    """

    default_message = _("Invalid value for parameter {param}.")

    def __init__(self, message=None, *, param, **kwargs):
        super().__init__(message)
        self.kwargs = {**kwargs, "param": param}


class InvalidIdCode(InvalidParameter):
    """Invalid ID code (format or checksum don't match)"""

    default_message = _("Invalid ID code.")

    def __init__(self, message=None, *, param="id_code", **kwargs):
        super().__init__(message, param=param, **kwargs)


class UserNotRegistered(EsteidError):
    """User not registered in the service."""

    status = HTTPStatus.CONFLICT

    default_message = _("User not registered in the service.")


class CanceledByUser(EsteidError):
    """User canceled the operation."""

    status = HTTPStatus.CONFLICT

    default_message = _("The signing operation was canceled by the user.")


class UserTimeout(EsteidError):
    """Failed to get PIN code from user within the interval specified by the service.

    For Mobile ID, this is about 2 minutes.
    https://github.com/SK-EID/MID#338-session-end-result-codes
    https://github.com/SK-EID/smart-id-documentation#5-session-end-result-codes
    """

    status = HTTPStatus.CONFLICT

    default_message = _("The signing operation timed out.")


class AlreadySignedByUser(InvalidParameters):
    """The container has been already signed by the same party.

    The class is based on InvalidParameters so that such errors are not logged.
    """

    status = HTTPStatus.CONFLICT

    default_message = _("The container has been already signed by the user.")



from unittest.mock import patch

import pytest

from esteid.exceptions import EsteidError
from esteid.signing import Signer


@patch.object(Signer, "SIGNING_METHODS", {})
def test_signer_register_subclass():
    def create_signer():
        class TestSigner(Signer):
            pass

        return TestSigner

    signer_class = create_signer()

    assert Signer.SIGNING_METHODS == {"test": signer_class}

    # Asserts that the signer class has not been registered yet, which is not true
    with pytest.raises(AssertionError):
        create_signer()


@patch.object(Signer, "SIGNING_METHODS", {})
def test_signer_select():
    class MySigner(Signer):
        pass

    assert Signer.select_signer("my") is MySigner
    with pytest.raises(EsteidError):
        Signer.select_signer("nonexistent")



from datetime import datetime
from typing import List, TYPE_CHECKING, Union

import attr
import pytz
from oscrypto.asymmetric import load_certificate


if TYPE_CHECKING:
    from oscrypto.asymmetric import Certificate as OsCryptoCertificate
    from asn1crypto.cms import Certificate as Asn1CryptoCertificate

import pyasice

from esteid.util import (
    camel_2_py,
    convert_status,
    convert_time,
    get_instance_converter,
    get_name_from_legacy_common_name,
    get_typed_list_converter,
    get_typed_list_validator,
)


class FromDictMixin(object):
    @classmethod
    def from_dict(cls, the_dict):
        return cls(**camel_2_py(the_dict))


@attr.s
class CertificatePolicy(FromDictMixin):
    oid = attr.ib()
    url = attr.ib(default=None)
    description = attr.ib(default=None)


@attr.s
class Certificate(FromDictMixin):
    issuer = attr.ib()
    issuer_serial = attr.ib()
    subject = attr.ib()
    valid_from: datetime = attr.ib(converter=convert_time)
    valid_to: datetime = attr.ib(converter=convert_time)

    policies = attr.ib(
        default=attr.Factory(list),
        validator=[
            attr.validators.instance_of(list),
            get_typed_list_validator(CertificatePolicy),
        ],
        converter=get_typed_list_converter(CertificatePolicy),
    )

    @classmethod
    def from_certificate(cls, cert: "Union[bytes, Asn1CryptoCertificate, OsCryptoCertificate]"):
        if isinstance(cert, bytes):
            cert = load_certificate(cert)
        cert_asn1: "Asn1CryptoCertificate" = getattr(cert, "asn1", cert)
        personal = cert_asn1.subject.native
        issuer = cert_asn1["tbs_certificate"]["issuer"]
        serial = cert_asn1["tbs_certificate"]["serial_number"].native

        validity = cert_asn1["tbs_certificate"]["validity"].native
        valid_from: datetime = validity["not_before"]
        valid_to: datetime = validity["not_after"]

        return cls(
            issuer=issuer.human_friendly,
            issuer_serial=str(serial),
            subject=personal["common_name"],
            valid_from=valid_from.replace(microsecond=0).astimezone(pytz.utc),
            valid_to=valid_to.replace(microsecond=0).astimezone(pytz.utc),
        )


@attr.s
class ResponderCertificate(Certificate):
    pass


@attr.s
class Signer(FromDictMixin):
    id_code = attr.ib()

    # Note: This should also support common_name argument coming in
    full_name = attr.ib()

    certificate: Certificate = attr.ib(
        validator=attr.validators.instance_of(Certificate), converter=get_instance_converter(Certificate)
    )

    @staticmethod
    def prepare_kwargs(kwargs):
        kwargs = camel_2_py(kwargs)
        full_name = kwargs.get("full_name", None)

        # If the full_name is None, check common_name field and use that
        if full_name is None:
            full_name = get_name_from_legacy_common_name(kwargs.pop("common_name", None))

        kwargs["full_name"] = full_name

        return kwargs

    @classmethod
    def from_certificate(cls, cert: "Union[bytes, Asn1CryptoCertificate, OsCryptoCertificate]"):
        """
        Get personal info from an oscrypto/asn1crypto Certificate object

        For a closer look at where the attributes come from:
        asn1crypto.x509.NameType
        """
        if isinstance(cert, bytes):
            cert = load_certificate(cert)
        cert: "Asn1CryptoCertificate" = getattr(cert, "asn1", cert)
        subject = cert.subject.native

        # ID codes usually given as PNO{EE,LT,LV}-XXXXXX.
        # LV ID codes contain a dash so we need to be careful about it.
        id_code = subject["serial_number"]
        if id_code.startswith("PNO"):
            prefix, id_code = id_code.split("-", 1)

        given_name = subject["given_name"]
        surname = subject["surname"]
        return cls(
            certificate=Certificate.from_certificate(cert),
            full_name=f"{given_name} {surname}",
            id_code=id_code,
        )


@attr.s
class Confirmation(FromDictMixin):
    produced_at = attr.ib()  # this might be a timestamp
    responder_id = attr.ib()

    responder_certificate = attr.ib(
        validator=attr.validators.instance_of(ResponderCertificate),
        converter=get_instance_converter(ResponderCertificate),
    )


@attr.s
class SignatureProductionPlace(FromDictMixin):
    postal_code = attr.ib()
    country_name = attr.ib()
    city = attr.ib()
    state_or_province = attr.ib()


@attr.s
class SignerRole(FromDictMixin):
    role = attr.ib()
    certified = attr.ib()


@attr.s
class SignatureInfo(FromDictMixin):
    signing_time = attr.ib(validator=attr.validators.instance_of(datetime), converter=convert_time)
    status = attr.ib(converter=convert_status)
    id = attr.ib()

    signer: Signer = attr.ib(
        validator=attr.validators.instance_of(Signer),
        converter=get_instance_converter(Signer, prepare_kwargs=Signer.prepare_kwargs),
    )
    confirmation: Confirmation = attr.ib(
        validator=attr.validators.instance_of(Confirmation), converter=get_instance_converter(Confirmation)
    )

    signature_production_place = attr.ib(
        validator=attr.validators.optional(attr.validators.instance_of(SignatureProductionPlace)),
        converter=attr.converters.optional(get_instance_converter(SignatureProductionPlace)),
        default=None,
    )

    signer_role = attr.ib(
        default=attr.Factory(list),
        validator=[
            attr.validators.instance_of(list),
            get_typed_list_validator(SignerRole),
        ],
        converter=get_typed_list_converter(SignerRole),
    )

    error = attr.ib(default=None)
    crl_info = attr.ib(default=None)

    # references Timestamps object in the docs, currently we do not parse it to a internal structure
    # http://sk-eid.github.io/dds-documentation/api/api_docs/#signeddocinfo
    timestamps = attr.ib(default=None)


@attr.s
class DataFileInfo(FromDictMixin):
    size = attr.ib()
    id = attr.ib()
    mime_type = attr.ib()
    content_type = attr.ib()
    digest_value = attr.ib()
    digest_type = attr.ib()
    filename = attr.ib()

    # Note: This is not documented as of 19:35 20.11.2017
    attributes = attr.ib(default=None)


@attr.s
class SignedDocInfo(FromDictMixin):
    format = attr.ib(default="BDOC")
    version = attr.ib(default="2.1")
    mime_type = attr.ib(default=pyasice.Container.MIME_TYPE)

    data_file_info: List[DataFileInfo] = attr.ib(
        default=attr.Factory(list),
        validator=[
            attr.validators.instance_of(list),
            get_typed_list_validator(DataFileInfo),
        ],
        converter=get_typed_list_converter(DataFileInfo),
    )

    signature_info: List[SignatureInfo] = attr.ib(
        default=attr.Factory(list),
        validator=[
            attr.validators.instance_of(list),
            get_typed_list_validator(SignatureInfo),
        ],
        converter=get_typed_list_converter(SignatureInfo),
    )


class DataFile:
    def __init__(self, file_name, mimetype, content_type, size, content, info=None):
        self.file_name = file_name
        self.mimetype = mimetype
        self.content_type = content_type
        self.size = size
        self.content = content

        self.info = info



import logging
import warnings

from esteid_certificates import get_certificate_file_name

from esteid import settings
from esteid.util import parse_legacy_dn, parse_rfc_dn

from .ocsp import OCSPVerifier


try:
    from django.utils.deprecation import MiddlewareMixin
except ImportError:  # Django < 1.10
    # Works perfectly for everyone using MIDDLEWARE_CLASSES
    MiddlewareMixin = object


logger = logging.getLogger(__name__)


ERROR_NO_DN = 3000
ERROR_INVALID_DN = 3001


class BaseIdCardMiddleware(MiddlewareMixin):
    """Base class for receiving authentication info from ID card.

    Note that this requires cooperation from the web server (eg Apache / Nginx) - the server must request client
    certificate and pass it on in specific HTTP headers.
    You should use one of the subclasses that's preconfigured for a specific server.

    The middleware also performs certificate validation check via OCSP if certificate and issuer headers are known and
    specified.

    If authentication information is present and valid, it's added to the request's `id_auth` attribute as dict
    containing at least the following items:

    - `serialNumber` - personal identification code, e.g. `37504170511`
    - `C` - country code, e.g. `EE`
    - `CN` - full name and personal id code, e.g. `SMITH,JOHN,37504170511`
    - `GN` - first name (given name), e.g. 'JOHN'
    - `SN` - last name (surname), e.g. 'SMITH'

    If authentication fails, the request will have `id_err` attribute with value of the error code.

    Documentation (In Estonian):
    https://eid.eesti.ee/index.php/Kasutaja_tuvastamine_veebis#Veebiserveri_konfigureerimine
    """

    DN_HEADER_NAME = None
    CERTIFICATE_HEADER_NAME = None
    ISSUER_HEADER_NAME = None

    # If any of these is not present in client DN, the authentication fails
    REQUIRED_DN_FIELDS = {
        "serialNumber",
        "C",
        "CN",
        "GN",
        "SN",
    }

    def process_request(self, request):
        assert self.DN_HEADER_NAME, "%s doesn't set DN_HEADER_NAME attribute" % self.__class__.__name__

        # Get the subject DN from header
        client_dn = request.META.get(self.DN_HEADER_NAME, None)
        if client_dn is None:
            setattr(request, "id_err", ERROR_NO_DN)
            return

        id_auth = self.parse_client_dn(client_dn)
        if not self.REQUIRED_DN_FIELDS.issubset(id_auth.keys()):
            logger.info("BaseIdCardMiddleware: required DN fields missing, client DN is %s", client_dn)
            setattr(request, "id_err", ERROR_INVALID_DN)
            return

        setattr(request, "id_auth", id_auth)

        if self.CERTIFICATE_HEADER_NAME is not None and self.ISSUER_HEADER_NAME is not None:
            issuer = request.META.get(self.ISSUER_HEADER_NAME, None)
            certificate = request.META.get(self.CERTIFICATE_HEADER_NAME, None)
            if issuer and certificate:
                certificate = self.prepare_certificate(certificate)
                setattr(request, "id_cert", certificate)

                # OCSP check
                issuer = self.parse_client_dn(issuer)
                response = self.verify_ocsp(certificate, issuer)

                # Code 0 means success, everything else is error
                if response != 0:
                    logger.info("BaseIdCardMiddleware: OCSP verification returned %s", response)
                    delattr(request, "id_auth")
                    # Pad OCSP error codes with 4000
                    setattr(request, "id_err", 4000 + response)

    def prepare_certificate(self, certificate):
        """ Use this in case the certificate value in HTTP header needs formatting. """
        return certificate

    def parse_client_dn(self, dn):
        """ Use either parse_legacy_dn or parse_rfc_dn depending on the format your proxy server users. """
        return parse_legacy_dn(dn)

    @classmethod
    def get_ocsp_url(cls):
        """Get ocsp responder certificate path

        Note: This is a separate method to allow easier overwriting via subclassing

        :return:str
        """
        return settings.OCSP_URL

    @classmethod
    def get_ocsp_responder_certificate_path(cls):
        """Get ocsp responder certificate path

        Note: This is a separate method to allow easier overwriting via subclassing

        :return:str
        """
        certificate_path = getattr(settings, "ESTEID_OCSP_RESPONDER_CERTIFICATE_PATH", None)
        test_cert_name = "TEST of SK OCSP RESPONDER 2011"
        live_cert_name = "sk-ocsp-responder-certificates"
        # Preserve compatibility
        if certificate_path == "TEST_of_SK_OCSP_RESPONDER_2011.pem":
            return get_certificate_file_name(test_cert_name)
        if certificate_path == "sk-ocsp-responder-certificates.pem":
            return get_certificate_file_name(live_cert_name)

        return test_cert_name if settings.ESTEID_DEMO else live_cert_name

    @classmethod
    def verify_ocsp(cls, certificate, issuer):
        """Runs OCSP verification and returns error code - 0 means success"""

        return OCSPVerifier(certificate, issuer, cls.get_ocsp_url(), cls.get_ocsp_responder_certificate_path()).verify()


class ApacheIdCardMiddleware(BaseIdCardMiddleware):
    """ID card middleware configured for Apache server

    Requires standard SSL_CLIENT_S_DN, SSL_CLIENT_I_DN and SSL_CLIENT_CERT headers.
    """

    DN_HEADER_NAME = "HTTP_SSL_CLIENT_S_DN"
    ISSUER_HEADER_NAME = "HTTP_SSL_CLIENT_I_DN"
    CERTIFICATE_HEADER_NAME = "HTTP_SSL_CLIENT_CERT"

    def prepare_certificate(self, certificate):
        return certificate.replace(" ", "\n").replace("\nCERTIFICATE-----", " CERTIFICATE-----")

    def parse_client_dn(self, dn):
        return parse_rfc_dn(dn)


class NginxIdCardMiddleware(BaseIdCardMiddleware):
    """ID card middleware configured for Nginx server

    Note that Nginx middleware does NOT perform validation (OCSP) checks for the received certificate.
    """

    DN_HEADER_NAME = "HTTP_X_CLIENT"


class IdCardMiddleware(NginxIdCardMiddleware):
    """ Deprecated - use ApacheIdCardMiddleware or NginxIdCardMiddleware """

    def __init__(self, *args, **kwargs):
        warnings.warn(
            "IdCardMiddleware is deprecated - use ApacheIdCardMiddleware or NginxIdCardMiddleware",
            DeprecationWarning,
            stacklevel=2,
        )
        super().__init__(*args, **kwargs)



import base64
import os
from pathlib import PurePath
from typing import Union

from django.core.files import File


class DataFile:
    """
    Interface for file objects that are added to signed containers.

    Constructor accepts a path to file (a string or a pathlib *Path) or a django File.

    Additional arguments include:
    - `mime_type`, defaults to "application/octet-stream";
    - `content` if for any reason the wrapped file's content should be ignored (it won't be read in this case);
    - `file_name`: a custom (base)name of the file as it would appear in the container.

    The container cares about 3 attributes:
    * file name (base name only)
        - taken from
    * mime type
    * the content, for calculating hash; `content` is obtained through a call to `read()`.
    """

    def __init__(
        self, wraps: Union[str, File, PurePath], mime_type: str = None, content: bytes = None, file_name: str = None
    ):
        if not isinstance(wraps, (str, File, PurePath)):
            raise TypeError(f"Invalid argument. Expected a Django File or path to file, got {type(wraps).__name__}")

        self.wrapped_file = wraps
        self.file_name = file_name or (os.path.basename(wraps if isinstance(wraps, str) else wraps.name))
        self.content = content

        # Can use the content_type attribute for UploadedFile
        self.mime_type = mime_type or getattr(wraps, "content_type", "application/octet-stream")

    def read(self):
        if self.content is None:
            if isinstance(self.wrapped_file, (str, PurePath)):
                with open(self.wrapped_file, "rb") as f:
                    self.content = f.read()
            else:
                # ... This doesn't work in Django 1.11 because the `open()` method returns None in that version.
                # with self.wrapped_file.open("rb") as f:
                with self.wrapped_file as f:
                    f.open("rb")
                    self.content = f.read()
        return self.content


class PredictableDict(dict):
    """
    Allows attribute-style access to values of a dict.

    Define necessary attributes as type annotations on your subclass:

        class Z(PredictableDict):
            required_attr: str
            optional_attr: Optional[str]

    and you will get nice attribute-style access with type hints.

    Validate the presence of all required attributes and type-check all attributes with:

        Z(required_attr="test").is_valid()

    Subclasses inherit annotated attributes and can override them, just like normal class attributes.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__dict__ = self

    def is_valid(self, raise_exception=True):
        for attr_name, attr_type in self._get_annotations().items():
            # type -> (type, )
            # Union[T, S] -> U.__args__ == (T, S)
            valid_types = getattr(attr_type, "__args__", (attr_type,))

            try:
                val = self[attr_name]
            except KeyError as e:
                # No error for optional fields.
                if type(None) in valid_types:
                    # Optional[T] == Union[T, NoneType]
                    continue

                if not raise_exception:
                    return False
                raise ValueError(f"Missing required key {attr_name}") from e

            if type(val) not in valid_types:
                if not raise_exception:
                    return False
                raise ValueError(f"Wrong type {type(val)} for key {attr_name}")

        return True

    @classmethod
    def _get_annotations(cls):
        """Collects annotations from all parent classes according to inheritance rules."""
        annotations = {}
        for klass in reversed(cls.__mro__):
            overrides = getattr(klass, "__annotations__", None)
            if overrides:
                annotations.update(overrides)
        return annotations


class InterimSessionData(PredictableDict):
    """
    Wrapper for temporary data stored between container preparation and finalization requests
    """

    digest_b64: str  # stores digest as a base64 encoded string, to allow JSON serialization
    temp_container_file: str
    temp_signature_file: str
    timestamp: int

    @property
    def digest(self):
        return base64.b64decode(self.digest_b64)

    @digest.setter
    def digest(self, value: bytes):
        self.digest_b64 = base64.b64encode(value).decode()



import base64
import hashlib

import pytest

from esteid.constants import HASH_SHA512, MOBILE_ID_DEMO_SERVICE_NAME, MOBILE_ID_DEMO_SERVICE_UUID, MOBILE_ID_DEMO_URL

from ..base import MobileIDService
from ..constants import EndResults
from ..i18n import TranslatedMobileIDService
from ..types import AuthenticateResult
from ..utils import get_verification_code


@pytest.fixture()
def demo_mid_api():
    return MobileIDService(MOBILE_ID_DEMO_SERVICE_UUID, MOBILE_ID_DEMO_SERVICE_NAME, MOBILE_ID_DEMO_URL)


@pytest.fixture()
def i18n_demo_mid_api():
    return TranslatedMobileIDService(MOBILE_ID_DEMO_SERVICE_UUID, MOBILE_ID_DEMO_SERVICE_NAME, MOBILE_ID_DEMO_URL)


@pytest.fixture()
def MID_DEMO_PHONE_EE_OK():
    return "+37200000766"


@pytest.fixture()
def MID_DEMO_PIN_EE_OK():
    return "60001019906"


@pytest.fixture()
def mid_auth_result(static_random_text):
    digest = hashlib.sha512(static_random_text).digest()
    return AuthenticateResult(
        session_id="FAKE",
        digest=digest,
        hash_type=HASH_SHA512,
        verification_code=get_verification_code(digest),
    )


@pytest.fixture()
def mid_auth_status_response(static_signature, static_signature_algorithm, static_certificate):
    return {
        "state": MobileIDService.ProcessingStates.COMPLETE,
        "result": EndResults.OK,
        "signature": {
            "algorithm": static_signature_algorithm,
            "value": base64.b64encode(static_signature),
        },
        "cert": base64.b64encode(static_certificate),
        "time": "2019-07-23T10:52:20",
    }


@pytest.fixture()
def mid_sign_status_response(static_signature, static_certificate, static_signature_algorithm):
    return {
        "state": MobileIDService.ProcessingStates.COMPLETE,
        "result": EndResults.OK,
        "signature": {
            "algorithm": static_signature_algorithm,
            "value": base64.b64encode(static_signature),
        },
        "time": "2019-07-23T10:52:20",
    }



import json
import logging
from http import HTTPStatus
from typing import BinaryIO, Type, TYPE_CHECKING, Union

from django.core.exceptions import ValidationError as DjangoValidationError
from django.http import Http404, HttpRequest, JsonResponse, QueryDict
from django.utils.translation import gettext

import pyasice

from esteid import settings
from esteid.exceptions import ActionInProgress, AlreadySignedByUser, CanceledByUser, EsteidError, InvalidParameters
from esteid.types import Signer as SignerData

from .signer import Signer


try:
    from rest_framework.exceptions import ValidationError as DRFValidationError
except ImportError:
    # If rest framework is not installed, create a stub class so the isinstance check is always false
    class DRFValidationError:
        pass


if TYPE_CHECKING:
    # Make type checkers aware of request.session attribute which is missing on the HttpRequest class
    from django.contrib.sessions import base_session

    class RequestType(HttpRequest):
        session: base_session.AbstractBaseSession
        data: dict


logger = logging.getLogger(__name__)


class SignViewMixin:
    class Status:
        ERROR = "error"
        PENDING = "pending"
        SUCCESS = "success"

    # this comes from the `url()` definition as in `View.as_view(signing_method='...')`
    signing_method: str = None

    def get_container(self, *args, **kwargs) -> Union[str, BinaryIO]:
        """
        Returns (path to|file handle of) the container to sign, if it exists prior to signing

        A string value is treated as a file path.
        A file handle is anything that provides a `read()` method that returns bytes.
        """
        raise NotImplementedError

    def save_container(self, container: pyasice.Container, *args, **kwargs):
        """
        Receives a container instance, expected to save the file contents as appropriate.

        Example with Django's UploadedFile:

            from esteid.compat import container_info
            instance = self.get_object()
            # Be sure to call `container_info(container)` before `container.finalize()`
            instance.container_info = container_info(container)
            instance.container = UploadedFile(container.finalize(), "signed_document.doc", container.MIME_TYPE)
            # OR:
            # buffer = container.finalize().getbuffer()
            # instance.container = SimpleUploadedFile("signed_document.doc", buffer, container.MIME_TYPE)
            instance.save()

        """
        raise NotImplementedError

    def get_files_to_sign(self, *args, **kwargs) -> list:
        """
        Returns list of files to sign, unless there is a pre-created container
        """
        raise NotImplementedError

    def get_success_response(self, *args, **kwargs):
        """
        Generates a success response when the signing process is complete (in `finalize()`)

        Can be customized to return additional data, such as links to download the container
        """
        return JsonResponse({"status": self.Status.SUCCESS})

    def check_eligibility(self, signer: Signer, container: pyasice.Container = None):
        if container is None or settings.ESTEID_ALLOW_ONE_PARTY_SIGN_TWICE:
            return

        signatories = []
        for signature in container.iter_signatures():
            subject_cert = signature.get_certificate()
            signatory = SignerData.from_certificate(subject_cert)
            signatories.append(signatory.id_code)

        if signer.id_code in signatories:
            raise AlreadySignedByUser

    def select_signer_class(self) -> Type["Signer"]:
        return Signer.select_signer(self.signing_method)

    def start_session(self, request: "RequestType", *args, **kwargs):
        """
        Initiates a signing session
        """

        signer_class = self.select_signer_class()
        signer = signer_class.start_session(request.session, request.data)

        try:
            container = self.get_container(*args, **kwargs)
        except NotImplementedError:
            container = None

        if not container:
            files_to_sign = self.get_files_to_sign(*args, **kwargs)
        else:
            files_to_sign = None

        self.check_eligibility(signer, container)

        response_to_user = signer.prepare(container, files_to_sign)

        return JsonResponse({**response_to_user, "status": self.Status.SUCCESS})

    def finish_session(self, request: "RequestType", *args, **kwargs):
        """
        Checks the status of a signing session and attempts to finalize signing
        """
        signer_class = self.select_signer_class()
        signer = signer_class.load_session(request.session)

        do_cleanup = True

        try:
            container = signer.finalize(getattr(request, "data", None))
            self.save_container(container, *args, **kwargs)

        except ActionInProgress as e:
            do_cleanup = False
            return JsonResponse({"status": self.Status.PENDING, **e.data}, status=e.status)

        finally:
            if do_cleanup:
                signer.cleanup()

        return self.get_success_response(*args, **kwargs)

    def report_error(self, e: EsteidError):
        return JsonResponse({"status": self.Status.ERROR, **e.get_user_error()}, status=e.status)

    def handle_user_cancel(self):
        pass

    def handle_errors(self, e: Exception, stage="start"):
        if isinstance(e, EsteidError):
            if isinstance(e, CanceledByUser):
                self.handle_user_cancel()

            # Do not log user input related errors
            if not isinstance(e, InvalidParameters):
                logger.exception("Failed to %s signing session.", stage)
            return self.report_error(e)

        if isinstance(e, (Http404, DRFValidationError)):
            raise

        if isinstance(e, DjangoValidationError):
            return JsonResponse(
                {"status": self.Status.ERROR, "error": e.__class__.__name__, "message": str(e)},
                status=HTTPStatus.CONFLICT,
            )

        logger.exception("Failed to %s signing session.", stage)
        return JsonResponse(
            {"status": self.Status.ERROR, "error": "Internal error", "message": gettext("Internal server error")},
            status=HTTPStatus.INTERNAL_SERVER_ERROR,
        )

    def post(self, request, *args, **kwargs):
        """
        Handles session start requests
        """
        try:
            return self.start_session(request, *args, **kwargs)
        except Exception as e:
            return self.handle_errors(e, stage="start")

    def patch(self, request, *args, **kwargs):
        """
        Handles session finish requests
        """
        try:
            return self.finish_session(request, *args, **kwargs)
        except Exception as e:
            return self.handle_errors(e, stage="finish")


class SignViewRestMixin(SignViewMixin):
    """
    To be used with rest-framework's APIView.
    """


class SignViewDjangoMixin(SignViewMixin):
    """
    To be used with plain Django class-based views (No rest-framework).

    Adds `data` attribute to the request with the POST or JSON data.
    """

    def post(self, request: "RequestType", *args, **kwargs):
        """
        Handles session start requests
        """
        request.data = self.parse_request(request)
        return super().start_session(request, *args, **kwargs)

    def patch(self, request: "RequestType", *args, **kwargs):
        """
        Handles session finish requests
        """
        request.data = self.parse_request(request)
        return super().finish_session(request, *args, **kwargs)

    @staticmethod
    def parse_request(request):
        """
        Parses PATCH/POST request bodies as JSON or urlencoded, and assigns `request.data`.

        Rationale:
        * Compatibility with REST Framework.
        * Allow JSON.
        * Django's request.POST only works for POST, not PATCH etc.
        """
        try:
            if request.content_type == "application/x-www-form-urlencoded":
                return QueryDict(request.body).dict()
            if request.content_type == "application/json":
                data = json.loads(request.body)
                if isinstance(data, dict):
                    return data
                raise InvalidParameters("Failed to parse request data as dict")
        except InvalidParameters:
            raise
        except Exception as e:
            raise InvalidParameters(
                f"Failed to parse the request body according to content type {request.content_type}"
            ) from e
        raise InvalidParameters(f"Unsupported request content type {request.content_type}")



"""
This module contains all configuration settings for django-esteid.
"""
from django.conf import settings

from esteid import constants
from esteid.constants import Countries


# Demo mode: whether to use demo services by default, including OCSP and TSA.
ESTEID_DEMO = getattr(settings, "ESTEID_DEMO", True)

# *** ID Card ***

# Whether to use the ID card signing method
ID_CARD_ENABLED = getattr(settings, "ID_CARD_ENABLED", False)

# *** Mobile ID ***

MOBILE_ID_ENABLED = getattr(settings, "MOBILE_ID_ENABLED", False)
# Whether to use demo services and credentials for Mobile ID. Default to global demo mode
MOBILE_ID_TEST_MODE = getattr(settings, "MOBILE_ID_TEST_MODE", ESTEID_DEMO)
# MobileID Relying party name and UUID, for DEMO they are always the same so no need to explicitly set them
MOBILE_ID_SERVICE_NAME = getattr(
    settings, "MOBILE_ID_SERVICE_NAME", None if not MOBILE_ID_TEST_MODE else constants.MOBILE_ID_DEMO_SERVICE_NAME
)
MOBILE_ID_SERVICE_UUID = getattr(
    settings, "MOBILE_ID_SERVICE_UUID", None if not MOBILE_ID_TEST_MODE else constants.MOBILE_ID_DEMO_SERVICE_UUID
)
MOBILE_ID_API_ROOT = getattr(
    settings,
    "MOBILE_ID_API_ROOT",
    constants.MOBILE_ID_DEMO_URL if MOBILE_ID_TEST_MODE else constants.MOBILE_ID_LIVE_URL,
)

# *** Smart ID ***

SMART_ID_ENABLED = getattr(settings, "SMART_ID_ENABLED", False)
# Whether to use demo services and credentials for Smart ID. Default to global demo mode
SMART_ID_TEST_MODE = getattr(settings, "SMART_ID_TEST_MODE", ESTEID_DEMO)
# SmartID Relying party name and UUID, for DEMO they are always the same so no need to explicitly set them
SMART_ID_SERVICE_NAME = getattr(
    settings, "SMART_ID_SERVICE_NAME", None if not SMART_ID_TEST_MODE else constants.SMART_ID_DEMO_SERVICE_NAME
)
SMART_ID_SERVICE_UUID = getattr(
    settings, "SMART_ID_SERVICE_UUID", None if not SMART_ID_TEST_MODE else constants.SMART_ID_DEMO_SERVICE_UUID
)
SMART_ID_API_ROOT = getattr(
    settings, "SMART_ID_API_ROOT", constants.SMART_ID_DEMO_URL if SMART_ID_TEST_MODE else constants.SMART_ID_LIVE_URL
)

# The default country (mostly for SmartID)
ESTEID_COUNTRY = getattr(settings, "ESTEID_COUNTRY", Countries.ESTONIA)

# *** Signature validity services: OCSP, TSA ***

# Whether to generate an LT-TS profile ASiC-E container (involves a TimeStamping confirmation)
ESTEID_USE_LT_TS = getattr(settings, "ESTEID_USE_LT_TS", True)

# URLs for OCSP and TSA services
OCSP_URL = getattr(settings, "ESTEID_OCSP_URL", constants.OCSP_DEMO_URL if ESTEID_DEMO else constants.OCSP_LIVE_URL)
TSA_URL = getattr(settings, "ESTEID_TSA_URL", constants.TSA_DEMO_URL if ESTEID_DEMO else constants.TSA_LIVE_URL)

# Used exclusively by esteid.middleware.BaseIdCardMiddleware
ESTEID_OCSP_RESPONDER_CERTIFICATE_PATH = getattr(settings, "ESTEID_OCSP_RESPONDER_CERTIFICATE_PATH", None)

# *** Misc settings ***

# Whether one signatory can sign the same container more than once. Default to allow for demo, disallow for live
ESTEID_ALLOW_ONE_PARTY_SIGN_TWICE = getattr(settings, "ESTEID_ALLOW_ONE_PARTY_SIGN_TWICE", ESTEID_DEMO)



import hashlib

import pytest

from ..utils import get_verification_code


@pytest.mark.parametrize(
    "hash_value,expected_verification_code",
    [
        (b"\x00\x00", "0000"),
        (b"00", "1584"),
        (hashlib.sha256(b"test").digest(), "5000"),
        (hashlib.sha512(b"test").digest(), "7679"),
    ],
)
def test_mobileid_get_verification_code(expected_verification_code, hash_value):
    assert get_verification_code(hash_value) == expected_verification_code



import base64
import binascii
import json
import os
from unittest.mock import patch
from urllib.parse import urlencode

import pytest

from django.contrib.sessions.backends import db
from django.test import Client
from django.urls import reverse

from esteid.flowtest.views import SigningTestView
from esteid.signing import Signer


@pytest.fixture()
def datafiles():
    return {"test.txt": {"content": base64.b64encode(b"text").decode(), "content_type": "text/plain"}}


@pytest.mark.parametrize("urlconf", ("sign-my", "sign-rest-my"))
@pytest.mark.parametrize("content_type", ("application/json", "application/x-www-form-urlencoded"))
def test_my_signing_flow(urlconf, content_type, datafiles):
    assert "my" in Signer.SIGNING_METHODS

    url = reverse(urlconf)

    client = Client()

    session = {}

    # This is a SUPER obvious way to work with test sessions.
    with patch.object(db, "SessionStore", return_value=session):
        session["__ddoc_files"] = datafiles

        if content_type == "application/json":
            data = json.dumps({})
        else:
            data = None
        response = client.post(url, data, content_type)

        assert response.status_code == 200
        assert response.json() == {
            "status": SigningTestView.Status.SUCCESS,
            "verification_code": "1234",
        }
        assert session[Signer._SESSION_KEY]["digest_b64"] == base64.b64encode(b"test").decode()
        temp_container_file = session[Signer._SESSION_KEY]["temp_container_file"]
        assert os.path.exists(temp_container_file)

        response = client.patch(url, data, content_type)
        assert response.status_code == 200
        assert response.json() == {
            "status": SigningTestView.Status.SUCCESS,
        }

        assert Signer._SESSION_KEY not in session, "Failed to clean up session"
        assert not os.path.exists(temp_container_file), "Failed to clean up files"


@pytest.mark.parametrize("urlconf", ("sign-mypost", "sign-rest-mypost"))
@pytest.mark.parametrize("content_type", ("application/json", "application/x-www-form-urlencoded"))
def test_my_post_signing_flow(urlconf, content_type, datafiles):
    assert "mypost" in Signer.SIGNING_METHODS

    url = reverse(urlconf)

    client = Client()

    session = {}
    with patch.object(db, "SessionStore", return_value=session):
        session["__ddoc_files"] = datafiles

        data = {"certificate": binascii.b2a_hex(b"test").decode()}
        if content_type == "application/json":
            data = json.dumps(data)
        else:
            data = urlencode(data)

        response = client.post(url, data, content_type)

        assert response.status_code == 200
        assert response.json() == {
            "status": SigningTestView.Status.SUCCESS,
            "verification_code": "1234",
        }
        assert session[Signer._SESSION_KEY]["digest_b64"] == base64.b64encode(b"test").decode()
        temp_container_file = session[Signer._SESSION_KEY]["temp_container_file"]
        assert os.path.exists(temp_container_file)

        data = {"signature_value": binascii.b2a_hex(b"test").decode()}
        if content_type == "application/json":
            data = json.dumps(data)
        else:
            data = urlencode(data)

        response = client.patch(url, data, content_type)
        assert response.status_code == 200
        assert response.json() == {
            "status": SigningTestView.Status.SUCCESS,
        }
        assert Signer._SESSION_KEY not in session, "Failed to clean up session"
        assert not os.path.exists(temp_container_file), "Failed to clean up files"



import hashlib
import struct


def get_verification_code(hash_value):
    """Compute Smart-ID verification code from a hash

    Verification Code is computed with: `integer(SHA256(hash)[-2:-1]) mod 10000`

    1. Take SHA256 result of hash_value
    2. Extract 2 rightmost bytes from it
    3. Interpret them as a big-endian unsigned short
    4. Take the last 4 digits in decimal

    Note: SHA256 is always used, e.g. the algorithm used when generating the hash does not matter

    based on https://github.com/SK-EID/smart-id-documentation#612-computing-the-verification-code
    """
    digest = hashlib.sha256(hash_value).digest()

    return "{:04}".format(struct.unpack(">H", digest[-2:])[0] % 10000)



import os
from typing import List

from pyasice import Container

from esteid.types import DataFile


SESSION_KEY = "__esteid_session"


class EsteidSessionError(Exception):
    pass


def get_esteid_session(request):
    return request.session.get(SESSION_KEY, {})


def start_esteid_session(request):
    delete_esteid_session(request)
    return {}


def update_esteid_session(request, **kwargs):
    session = request.session
    data = session.get(SESSION_KEY, {})
    data.update(kwargs)
    session[SESSION_KEY] = data  # Django session object requires setting keys explicitly


def delete_esteid_session(request):
    session_data = request.session.pop(SESSION_KEY, {})

    temp_signature_file = session_data.get("temp_signature_file")
    if temp_signature_file:
        try:
            os.unlink(temp_signature_file)
        except OSError:
            pass

    temp_container_file = session_data.get("temp_container_file")
    if temp_container_file:
        try:
            os.unlink(temp_container_file)
        except OSError:
            pass


def open_container(container_path: str = None, files_to_sign: List[DataFile] = None):
    """Open an existing BDoc container, or create a container with files to sign"""
    if container_path:
        # Take files from an existing container, ignore files_to_sign
        container = Container.open(container_path)
    elif files_to_sign:
        container = Container()
        for data_file in files_to_sign:
            container.add_file(data_file.file_name, data_file.content, data_file.mimetype)
    else:
        raise ValueError("No files to sign")  # this is a 500 error because it's not the user's fault

    return container



import logging
from typing import List, Optional

import pyasice
from pyasice import Container, XmlSignature

from esteid.exceptions import ActionInProgress, InvalidIdCode, InvalidParameters
from esteid.signing import DataFile, Signer
from esteid.signing.types import InterimSessionData, PredictableDict

from .base import validate_id_code
from .constants import Countries
from .i18n import TranslatedSmartIDService


logger = logging.getLogger(__name__)


class UserInput(PredictableDict):
    id_code: str
    country: Optional[str]

    def is_valid(self, raise_exception=True):
        """
        Raises InvalidIdCode or ValueError
        """
        result = super().is_valid(raise_exception=raise_exception)
        if result:
            if not (self.get("country") and self.country in Countries.ALL):
                self.country = Countries.ESTONIA

            validate_id_code(self.id_code, self.country)

        return result


class SmartIdSessionData(InterimSessionData):
    session_id: str


class SmartIdSigner(Signer):
    id_code: str
    country: str

    SessionData = SmartIdSessionData
    session_data: SmartIdSessionData

    def setup(self, initial_data: dict = None):
        """
        Receives user input via POST: `id_code`, `country`
        """
        if not isinstance(initial_data, dict):
            raise InvalidParameters("Missing required parameters")

        user_input = UserInput(**initial_data)

        try:
            user_input.is_valid()
        except InvalidIdCode:
            # Just to be explicit
            raise
        except ValueError as e:
            raise InvalidParameters from e

        self.id_code = user_input.id_code
        self.country = user_input.country

    def prepare(self, container_file=None, files: List[DataFile] = None) -> dict:
        container = self.open_container(container_file, files)

        service = TranslatedSmartIDService.get_instance()

        certificate, document_number = service.select_signing_certificate(self.id_code, self.country)

        xml_sig = container.prepare_signature(certificate)

        sign_session = service.sign_by_document_number(document_number, xml_sig.signed_data())

        self.save_session_data(
            digest=sign_session.digest,
            container=container,
            xml_sig=xml_sig,
            session_id=sign_session.session_id,
        )

        return {
            "verification_code": sign_session.verification_code,
        }

    def finalize(self, data: dict = None) -> Container:
        digest = self.session_data.digest
        session_id = self.session_data.session_id

        temp_signature_file = self.session_data.temp_signature_file
        temp_container_file = self.session_data.temp_container_file

        with open(temp_signature_file, "rb") as f:
            xml_sig = pyasice.XmlSignature(f.read())

        service = TranslatedSmartIDService.get_instance()

        try:
            status = service.sign_status(session_id, digest)
        except ActionInProgress:
            # Just to be explicit about the InProgress status
            raise

        container = pyasice.Container.open(temp_container_file)

        xml_sig.set_signature_value(status.signature)

        self.finalize_xml_signature(xml_sig)

        container.add_signature(xml_sig)

        return container

    def save_session_data(self, *, digest: bytes, container: Container, xml_sig: XmlSignature, session_id: str):
        data_obj = self.session_data
        data_obj.session_id = session_id

        super().save_session_data(digest=digest, container=container, xml_sig=xml_sig)



# -*- coding: utf-8 -*-
import pytest

from esteid.util import get_id_from_legacy_common_name, get_name_from_legacy_common_name


@pytest.mark.parametrize(
    "common_name,expected_name",
    [
        ("TESTNUMBER,SEITSMES,51001091072", "Seitsmes Testnumber"),
        ("TESTNUMBER\\,SEITSMES\\,51001091072", "Seitsmes Testnumber"),
        ("TEST-NUMBER,SEITSMES MEES,51001091072", "Seitsmes Mees Test-Number"),
        ("O’CONNEŽ-ŠUSLIK,MARY ÄNN,11412090004", "Mary Änn O’Connež-Šuslik"),
    ],
)
def test_get_name_from_legacy_common_name(common_name, expected_name):
    result = get_name_from_legacy_common_name(common_name)

    assert result == expected_name


@pytest.mark.parametrize(
    "common_name,expected_id",
    [
        ("TESTNUMBER,SEITSMES,51001091072", "51001091072"),
        ("TESTNUMBER\\,SEITSMES\\,51001091072", "51001091072"),
        ("TEST-NUMBER,SEITSMES MEES,51001091072", "51001091072"),
        ("O’CONNEŽ-ŠUSLIK,MARY ÄNN,11412090004", "11412090004"),
    ],
)
def test_get_id_from_legacy_common_name(common_name, expected_id):
    result = get_id_from_legacy_common_name(common_name)

    assert result == expected_id



import json

import pytest

import attr
import oscrypto.asymmetric
from django.core.serializers.json import DjangoJSONEncoder

import pyasice

from esteid.compat import container_info
from esteid.types import Certificate, DataFileInfo, SignedDocInfo, Signer


def test_signeddocinfo_from_dict(signed_doc_dict):
    # smoketest for signed doc parsing
    instance = SignedDocInfo.from_dict(signed_doc_dict)

    assert instance.format == "BDOC"
    assert instance.version == "2.1"

    assert instance.data_file_info == [
        DataFileInfo.from_dict(
            {
                "contentType": "HASHCODE",
                "digestType": "sha256",
                "digestValue": "Mmb3rosrMXMY86F4P+Pj0VQEQnpDquhEarJ00RelWXE=",
                "filename": "something.pdf",
                "id": "something.pdf",
                "mime_type": "application/pdf",
                "size": 27216,
            }
        )
    ]


def test_signeddocinfo_from_container(signed_container_file):
    with pyasice.Container(signed_container_file) as container:
        data = container_info(container)

    assert data.format == "BDOC"
    assert data.version == "2.1"
    assert data.mime_type == pyasice.Container.MIME_TYPE

    assert len(data.signature_info) == 1
    assert data.signature_info[0].signing_time
    assert data.signature_info[0].signer.id_code == "60001019906"

    assert len(data.data_file_info) == 1
    assert data.data_file_info[0].filename == "test.txt"


@pytest.mark.parametrize("cert_type", ["bytes", "asn1", "oscrypto"])
def test_certificate_from_cert(static_certificate, cert_type):
    if cert_type == "bytes":
        cert = static_certificate
    else:
        cert = oscrypto.asymmetric.load_certificate(static_certificate)
        if cert_type == "asn1":
            cert = cert.asn1

    cert_data = Certificate.from_certificate(cert)
    assert cert_data.issuer_serial == "116050271893176812114901422365303754679"
    assert cert_data.issuer == (
        "Common Name: TEST of NQ-SK 2016, "
        "Organization Identifier: NTREE-10747013, "
        "Organization: AS Sertifitseerimiskeskus, Country: EE"
    )
    assert cert_data.valid_from.isoformat() == "2017-02-02T09:14:37+00:00"
    assert cert_data.valid_to.isoformat() == "2022-02-01T21:59:59+00:00"
    assert cert_data.subject == "SMART-ID,HELLO,PNOEE-11702020200"


@pytest.mark.parametrize("cert_type", ["bytes", "asn1", "oscrypto"])
def test_signer_data_from_oscrypto_cert(static_certificate, cert_type):
    if cert_type == "bytes":
        cert = static_certificate
    else:
        cert = oscrypto.asymmetric.load_certificate(static_certificate)
        if cert_type == "asn1":
            cert = cert.asn1

    signer_data = Signer.from_certificate(cert)
    assert signer_data.full_name == "HELLO SMART-ID"
    assert signer_data.id_code == "11702020200"
    assert isinstance(signer_data.certificate, Certificate)


def test_signer_data_to_json(static_certificate):
    signer_data = Signer.from_certificate(static_certificate)

    output = json.dumps(attr.asdict(signer_data), cls=DjangoJSONEncoder)
    assert Signer.from_dict(json.loads(output)) == signer_data



from django.conf.urls import url

from esteid.signing import Signer

from .views import SigningTestRestView, SigningTestView


urlpatterns = [
    url(f"^sign/{method}/", SigningTestView.as_view(signing_method=method), name=f"sign-{method}")
    for method in Signer.SIGNING_METHODS
]

urlpatterns += [
    url(f"^sign-rest/{method}/", SigningTestRestView.as_view(signing_method=method), name=f"sign-rest-{method}")
    for method in Signer.SIGNING_METHODS
]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import esteid


try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

version = esteid.__version__

readme = open("README.md").read()
requirements_base = open("requirements-base.txt").readlines()
requirements = open("requirements.txt").readlines()

setup(
    name="django-esteid",
    version=version,
    description="""Django-esteid is a package that provides Esteid based authentication for your Django applications.""",
    long_description=readme,
    long_description_content_type="text/markdown",
    author="Thorgate",
    author_email="jyrno@thorgate.eu",
    url="https://github.com/thorgate/django-esteid",
    packages=[
        "esteid",
    ],
    include_package_data=True,
    install_requires=[line for line in requirements + requirements_base if line and not line.startswith(("#", "-"))],
    license="BSD",
    zip_safe=False,
    keywords="esteid django",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Framework :: Django",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
)



def get_verification_code(hash_value):
    """Compute Mobile-ID verification code from a hash

    Excerpt from https://github.com/SK-EID/MID#241-verification-code-calculation-algorithm

    1. 6 bits from the beginning of hash and 7 bits from the end of hash are taken.
    2. The resulting 13 bits are transformed into decimal number and printed out.

    The Verification code is a decimal 4-digits number in range 0000...8192, always 4 digits are displayed (e.g. 0041).

    :param bytes hash_value:
    """
    if not isinstance(hash_value, bytes):
        raise TypeError(f"Invalid hash value: expected bytes, got {type(hash_value)}")

    if not hash_value:
        raise ValueError("Hash value can not be empty")

    leading_byte = hash_value[0]
    trailing_byte = hash_value[-1]

    leading_6_bits = leading_byte >> 2
    trailing_7_bits = trailing_byte & 0x7F

    return "{:04d}".format((leading_6_bits << 7) + trailing_7_bits)



import base64

import pytest

from ...constants import HASH_SHA512, SMART_ID_DEMO_SERVICE_NAME, SMART_ID_DEMO_SERVICE_UUID, SMART_ID_DEMO_URL
from ...util import generate_hash
from ..base import SmartIDService
from ..constants import CERTIFICATE_LEVEL_QUALIFIED, EndResults
from ..i18n import TranslatedSmartIDService
from ..types import AuthenticateResult
from ..utils import get_verification_code


@pytest.fixture
def demo_api():
    return SmartIDService(SMART_ID_DEMO_SERVICE_UUID, SMART_ID_DEMO_SERVICE_NAME, SMART_ID_DEMO_URL)


@pytest.fixture
def i18n_demo_api():
    return TranslatedSmartIDService(SMART_ID_DEMO_SERVICE_UUID, SMART_ID_DEMO_SERVICE_NAME, SMART_ID_DEMO_URL)


@pytest.fixture
def SMARTID_DEMO_ID_CODE_EE():
    return "10101010005"


@pytest.fixture
def SMARTID_DEMO_ID_CODE_LT():
    return "10101010005"


@pytest.fixture
def SMARTID_DEMO_ID_CODE_LV():
    return "010101-10006"


@pytest.fixture
def static_auth_result(static_random_text):
    return AuthenticateResult(
        session_id="FAKE",
        hash_value=generate_hash(HASH_SHA512, static_random_text),
        hash_type=HASH_SHA512,
        verification_code=get_verification_code(static_random_text),
    )


@pytest.fixture
def static_status_response(static_signature, static_signature_algorithm, static_certificate):
    return {
        "state": SmartIDService.ProcessingStates.COMPLETE,
        "result": {
            "endResult": EndResults.OK,
            "documentNumber": "$documentNumber$",
        },
        "signature": {
            "algorithm": static_signature_algorithm,
            "value": base64.b64encode(static_signature),
        },
        "cert": {
            "value": base64.b64encode(static_certificate),
            "certificateLevel": CERTIFICATE_LEVEL_QUALIFIED,
        },
    }



import importlib

from django.test import override_settings

import esteid.settings


def test_importlib():
    with override_settings(MOBILE_ID_TEST_MODE=False):
        importlib.reload(esteid.settings)
        assert esteid.settings.MOBILE_ID_TEST_MODE is False



from esteid.constants import Countries  # noqa


CERTIFICATE_LEVEL_QUALIFIED = "QUALIFIED"
CERTIFICATE_LEVEL_ADVANCED = "ADVANCED"

CERTIFICATE_LEVELS = (
    CERTIFICATE_LEVEL_QUALIFIED,
    CERTIFICATE_LEVEL_ADVANCED,
)


class EndResults:
    """
    https://github.com/SK-EID/smart-id-documentation#5-session-end-result-codes
    """

    OK = "OK"  # session was completed successfully
    USER_REFUSED = "USER_REFUSED"  # user refused the session
    TIMEOUT = "TIMEOUT"  # there was a timeout, i.e. end user did not confirm or he operation within given time-frame
    DOCUMENT_UNUSABLE = "DOCUMENT_UNUSABLE"  # for some reason, this RP request cannot be completed

    ALL = (OK, DOCUMENT_UNUSABLE, TIMEOUT, USER_REFUSED)
